import contextlib
import json
import os
import pprint
import shutil
import sys

from eth_utils import (
    is_dict,
    is_same_address,
)

import common
from web3 import Web3
from web3.utils.module_testing.emitter_contract import (
    EMITTER_ABI,
    EMITTER_BYTECODE,
    EMITTER_ENUM,
)
from web3.utils.module_testing.math_contract import (
    MATH_ABI,
    MATH_BYTECODE,
)
from web3.utils.toolz import (
    merge,
)


def generate_go_ethereum_fixture(destination_dir):
    with contextlib.ExitStack() as stack:
        datadir = stack.enter_context(common.tempdir())

        keystore_dir = os.path.join(datadir, 'keystore')
        common.ensure_path_exists(keystore_dir)
        keyfile_path = os.path.join(keystore_dir, common.KEYFILE_FILENAME)
        with open(keyfile_path, 'w') as keyfile:
            keyfile.write(common.KEYFILE_DATA)
        genesis_file_path = os.path.join(datadir, 'genesis.json')
        with open(genesis_file_path, 'w') as genesis_file:
            genesis_file.write(json.dumps(common.GENESIS_DATA))

        geth_ipc_path_dir = stack.enter_context(common.tempdir())
        geth_ipc_path = os.path.join(geth_ipc_path_dir, 'geth.ipc')

        geth_port = common.get_open_port()
        geth_binary = common.get_geth_binary()

        geth_proc = stack.enter_context(common.get_geth_process(  # noqa: F841
            geth_binary=geth_binary,
            datadir=datadir,
            genesis_file_path=genesis_file_path,
            ipc_path=geth_ipc_path,
            port=geth_port,
            networkid=str(common.GENESIS_DATA['config']['chainId'])
        ))

        common.wait_for_socket(geth_ipc_path)
        web3 = Web3(Web3.IPCProvider(geth_ipc_path))
        chain_data = setup_chain_state(web3)
        static_data = {
            'raw_txn_account': common.RAW_TXN_ACCOUNT,
            'keyfile_pw': common.KEYFILE_PW,
        }
        pprint.pprint(merge(chain_data, static_data))

        shutil.copytree(datadir, destination_dir)


def setup_chain_state(web3):
    coinbase = web3.eth.coinbase

    assert is_same_address(coinbase, common.COINBASE)

    #
    # Math Contract
    #
    math_contract_factory = web3.eth.contract(
        abi=MATH_ABI,
        bytecode=MATH_BYTECODE,
    )
    math_deploy_receipt = common.deploy_contract(web3, 'math', math_contract_factory)
    assert is_dict(math_deploy_receipt)

    #
    # Emitter Contract
    #
    emitter_contract_factory = web3.eth.contract(
        abi=EMITTER_ABI,
        bytecode=EMITTER_BYTECODE,
    )
    emitter_deploy_receipt = common.deploy_contract(web3, 'emitter', emitter_contract_factory)
    emitter_contract = emitter_contract_factory(emitter_deploy_receipt['contractAddress'])

    txn_hash_with_log = emitter_contract.functions.logDouble(
        which=EMITTER_ENUM['LogDoubleWithIndex'], arg0=12345, arg1=54321,
    ).transact({
        'from': web3.eth.coinbase,
    })
    print('TXN_HASH_WITH_LOG:', txn_hash_with_log)
    txn_receipt_with_log = common.mine_transaction_hash(web3, txn_hash_with_log)
    block_with_log = web3.eth.getBlock(txn_receipt_with_log['blockHash'])
    print('BLOCK_HASH_WITH_LOG:', block_with_log['hash'])

    #
    # Empty Block
    #
    empty_block_number = common.mine_block(web3)
    print('MINED_EMPTY_BLOCK')
    empty_block = web3.eth.getBlock(empty_block_number)
    assert is_dict(empty_block)
    assert not empty_block['transactions']
    print('EMPTY_BLOCK_HASH:', empty_block['hash'])

    #
    # Block with Transaction
    #
    web3.personal.unlockAccount(coinbase, common.KEYFILE_PW)
    web3.miner.start(1)
    mined_txn_hash = web3.eth.sendTransaction({
        'from': coinbase,
        'to': coinbase,
        'value': 1,
        'gas': 21000,
        'gas_price': web3.eth.gasPrice,
    })
    mined_txn_receipt = common.mine_transaction_hash(web3, mined_txn_hash)
    print('MINED_TXN_HASH:', mined_txn_hash)
    block_with_txn = web3.eth.getBlock(mined_txn_receipt['blockHash'])
    print('BLOCK_WITH_TXN_HASH:', block_with_txn['hash'])

    geth_fixture = {
        'math_deploy_txn_hash': math_deploy_receipt['transactionHash'],
        'math_address': math_deploy_receipt['contractAddress'],
        'emitter_deploy_txn_hash': emitter_deploy_receipt['transactionHash'],
        'emitter_address': emitter_deploy_receipt['contractAddress'],
        'txn_hash_with_log': txn_hash_with_log,
        'block_hash_with_log': block_with_log['hash'],
        'empty_block_hash': empty_block['hash'],
        'mined_txn_hash': mined_txn_hash,
        'block_with_txn_hash': block_with_txn['hash'],
    }
    return geth_fixture


if __name__ == '__main__':
    fixture_dir = sys.argv[1]
    generate_go_ethereum_fixture(fixture_dir)



import pytest

from websockets.exceptions import (
    ConnectionClosed,
)

from web3 import Web3


class MiscWebsocketTest:

    def test_websocket_max_size_error(self, web3, endpoint_uri):
        w3 = Web3(Web3.WebsocketProvider(
            endpoint_uri=endpoint_uri, websocket_kwargs={'max_size': 1})
        )
        with pytest.raises(ConnectionClosed):
            w3.eth.getBlock(0)



import pytest

from web3.utils.transactions import (
    prepare_replacement_transaction,
)

SIMPLE_CURRENT_TRANSACTION = {
    'blockHash': None,
    'hash': '0x0',
    'nonce': 2,
    'gasPrice': 10,
}


def test_prepare_transaction_replacement(web3):
    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 1,
        'nonce': 2,
    }
    replacement_transaction = prepare_replacement_transaction(
        web3, current_transaction, new_transaction)

    assert replacement_transaction == {
        'value': 1,
        'nonce': 2,
        'gasPrice': 11,
    }


def test_prepare_transaction_replacement_without_nonce_sets_correct_nonce(web3):
    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 1,
    }
    replacement_transaction = prepare_replacement_transaction(
        web3, current_transaction, new_transaction)
    assert replacement_transaction == {
        'value': 1,
        'nonce': 2,
        'gasPrice': 11,
    }


def test_prepare_transaction_replacement_already_mined_raises(web3):
    with pytest.raises(ValueError):
        prepare_replacement_transaction(
            web3, {'blockHash': '0xa1a1a1', 'hash': '0x0'}, {'value': 2})


def test_prepare_transaction_replacement_nonce_mismatch_raises(web3):
    with pytest.raises(ValueError):
        prepare_replacement_transaction(web3, {
            'blockHash': None,
            'hash': '0x0',
            'nonce': 1,
        }, {
            'nonce': 2,
        })


def test_prepare_transaction_replacement_not_higher_gas_price_raises(web3):
    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 1,
        'gasPrice': 5,
    }
    with pytest.raises(ValueError):
        prepare_replacement_transaction(
            web3, current_transaction, new_transaction)

    # Also raises when equal to the current transaction
    new_transaction['gasPrice'] = 10
    with pytest.raises(ValueError):
        prepare_replacement_transaction(web3, current_transaction, new_transaction)


def test_prepare_transaction_replacement_gas_price_defaulting(web3):
    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 2,
    }
    replacement_transaction = prepare_replacement_transaction(
        web3, current_transaction, new_transaction)

    assert replacement_transaction['gasPrice'] == 11


def test_prepare_transaction_replacement_gas_price_defaulting_when_strategy_higer(web3):

    def higher_gas_price_strategy(web3, txn):
        return 20

    web3.eth.setGasPriceStrategy(higher_gas_price_strategy)

    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 2,
    }

    replacement_transaction = prepare_replacement_transaction(
        web3, current_transaction, new_transaction)

    assert replacement_transaction['gasPrice'] == 20


def test_prepare_transaction_replacement_gas_price_defaulting_when_strategy_lower(web3):

    def lower_gas_price_strategy(web3, txn):
        return 5

    web3.eth.setGasPriceStrategy(lower_gas_price_strategy)

    current_transaction = SIMPLE_CURRENT_TRANSACTION
    new_transaction = {
        'value': 2,
    }

    replacement_transaction = prepare_replacement_transaction(
        web3, current_transaction, new_transaction)

    assert replacement_transaction['gasPrice'] == 11



import pytest

from web3.manager import (
    RequestManager,
)
from web3.providers import (
    BaseProvider,
)


def test_provider_property_setter_and_getter(middleware_factory):
    provider = BaseProvider()

    middleware_a = middleware_factory()
    middleware_b = middleware_factory()
    middleware_c = middleware_factory()
    assert middleware_a is not middleware_b
    assert middleware_a is not middleware_c

    manager = RequestManager(None, provider, middlewares=[])

    assert tuple(manager.middleware_stack) == tuple()

    manager.middleware_stack.add(middleware_a)
    manager.middleware_stack.add(middleware_b)

    manager.middleware_stack.clear()

    assert tuple(manager.middleware_stack) == tuple()

    manager.middleware_stack.add(middleware_c)
    manager.middleware_stack.add(middleware_b)
    manager.middleware_stack.add(middleware_a)

    with pytest.raises(ValueError):
        manager.middleware_stack.add(middleware_b)

    assert tuple(manager.middleware_stack) == (
        middleware_a,
        middleware_b,
        middleware_c,
    )


def test_add_named_middleware(middleware_factory):
    mw = middleware_factory()
    manager = RequestManager(None, BaseProvider(), middlewares=[(mw, 'the-name')])
    assert len(manager.middleware_stack) == 1

    assert tuple(manager.middleware_stack) == (mw, )


def test_add_named_duplicate_middleware(middleware_factory):
    mw = middleware_factory()
    manager = RequestManager(None, BaseProvider(), middlewares=[(mw, 'the-name'), (mw, 'name2')])
    assert tuple(manager.middleware_stack) == (mw, mw)

    manager.middleware_stack.clear()
    assert len(manager.middleware_stack) == 0

    manager.middleware_stack.add(mw, 'name1')
    manager.middleware_stack.add(mw, 'name2')
    assert tuple(manager.middleware_stack) == (mw, mw)


def test_add_duplicate_middleware(middleware_factory):
    mw = middleware_factory()
    with pytest.raises(ValueError):
        manager = RequestManager(None, BaseProvider(), middlewares=[mw, mw])

    manager = RequestManager(None, BaseProvider(), middlewares=[])
    manager.middleware_stack.add(mw)

    with pytest.raises(ValueError):
        manager.middleware_stack.add(mw)
    assert tuple(manager.middleware_stack) == (mw, )


def test_replace_middleware(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, (mw2, '2nd'), mw3])

    assert tuple(manager.middleware_stack) == (mw1, mw2, mw3)

    mw_replacement = middleware_factory()
    manager.middleware_stack.replace('2nd', mw_replacement)

    assert tuple(manager.middleware_stack) == (mw1, mw_replacement, mw3)

    manager.middleware_stack.remove('2nd')

    assert tuple(manager.middleware_stack) == (mw1, mw3)


def test_replace_middleware_without_name(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, mw2, mw3])

    assert tuple(manager.middleware_stack) == (mw1, mw2, mw3)

    mw_replacement = middleware_factory()
    manager.middleware_stack.replace(mw2, mw_replacement)

    assert tuple(manager.middleware_stack) == (mw1, mw_replacement, mw3)

    manager.middleware_stack.remove(mw_replacement)

    assert tuple(manager.middleware_stack) == (mw1, mw3)


def test_add_middleware(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, mw2])

    manager.middleware_stack.add(mw3)

    assert tuple(manager.middleware_stack) == (mw3, mw1, mw2)


def test_bury_middleware(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, mw2])

    manager.middleware_stack.inject(mw3, layer=0)

    assert tuple(manager.middleware_stack) == (mw1, mw2, mw3)


def test_bury_named_middleware(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, mw2])

    manager.middleware_stack.inject(mw3, name='middleware3', layer=0)

    assert tuple(manager.middleware_stack) == (mw1, mw2, mw3)

    # make sure middleware was injected with correct name, by trying to remove
    # it by name.
    manager.middleware_stack.remove('middleware3')

    assert tuple(manager.middleware_stack) == (mw1, mw2)


def test_remove_middleware(middleware_factory):
    mw1 = middleware_factory()
    mw2 = middleware_factory()
    mw3 = middleware_factory()

    manager = RequestManager(None, BaseProvider(), middlewares=[mw1, mw2, mw3])

    assert tuple(manager.middleware_stack) == (mw1, mw2, mw3)

    manager.middleware_stack.remove(mw2)

    assert tuple(manager.middleware_stack) == (mw1, mw3)



import random

from eth_utils import (
    to_dict,
)
from hypothesis import (
    given,
    strategies as st,
)

from web3.utils.caching import (
    generate_cache_key,
)


@to_dict
def shuffle_dict(_dict):
    keys = list(_dict.keys())
    random.shuffle(keys)
    for key in keys:
        yield key, _dict[key]


def extend_fn(children):
    lists_st = st.lists(children)
    dicts_st = st.dictionaries(st.text(), children)
    return lists_st | dicts_st


all_st = st.recursive(
    st.none() | st.integers() | st.booleans() | st.floats() | st.text() | st.binary(),
    extend_fn,
)


def recursive_shuffle_dict(v):
    if isinstance(v, dict):
        return shuffle_dict(v)
    elif isinstance(v, (list, tuple)):
        return type(v)((recursive_shuffle_dict(_v) for _v in v))
    else:
        return v


@given(value=all_st)
def test_key_generation_is_deterministic(value):
    left = recursive_shuffle_dict(value)
    right = recursive_shuffle_dict(value)
    left_key = generate_cache_key(left)
    right_key = generate_cache_key(right)
    assert left_key == right_key




import pytest

from ens.main import (
    AddressMismatch,
    UnauthorizedError,
    UnownedName,
)
from web3 import Web3


'''
API at: https://github.com/carver/ens.py/issues/2
'''


@pytest.fixture()
def TEST_ADDRESS(address_conversion_func):
    return address_conversion_func("0x000000000000000000000000000000000000dEaD")


@pytest.mark.parametrize(
    'name, normalized_name, namehash_hex',
    [
        (
            'tester.eth',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        (
            'tester',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        (
            'TESTER',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        (
            'tester．eth',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        (
            'tester。eth',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        (
            'tester｡eth',
            'tester.eth',
            '2a7ac1c833d35677c2ff34a908951de142cc1653de6080ad4e38f4c9cc00aafe',
        ),
        # confirm that set-owner works
        (
            'lots.of.subdomains.tester.eth',
            'lots.of.subdomains.tester.eth',
            '0d62a759aa1f1c9680de8603a12a5eb175cd1bfa79426229868eba99f4dce692',
        ),
        (
            'lots.of.subdomains.tester',
            'lots.of.subdomains.tester.eth',
            '0d62a759aa1f1c9680de8603a12a5eb175cd1bfa79426229868eba99f4dce692',
        ),
    ],
)
def test_setup_name(ens, name, normalized_name, namehash_hex):
    address = ens.web3.eth.accounts[3]
    assert not ens.name(address)
    owner = ens.owner('tester')

    ens.setup_name(name, address)
    assert ens.name(address) == normalized_name

    # check that .eth is only appended if guess_tld is True
    if ens.nameprep(name) != normalized_name:
        assert ens.address(name, guess_tld=False) is None

    # check that the correct namehash is set:
    node = Web3.toBytes(hexstr=namehash_hex)
    assert ens.resolver(normalized_name).addr(node) == address

    # check that the correct owner is set:
    assert ens.owner(name) == owner

    ens.setup_name(None, address)
    ens.setup_address(name, None)
    assert not ens.name(address)
    assert not ens.address(name)


def test_cannot_set_name_on_mismatch_address(ens, TEST_ADDRESS):
    ens.setup_address('mismatch-reverse.tester.eth', TEST_ADDRESS)
    with pytest.raises(AddressMismatch):
        ens.setup_name('mismatch-reverse.tester.eth', '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413')


def test_setup_name_default_address(ens):
    name = 'reverse-defaults-to-forward.tester.eth'
    owner = ens.owner('tester.eth')
    new_resolution = ens.web3.eth.accounts[-1]
    ens.setup_address(name, new_resolution)
    assert not ens.name(new_resolution)
    assert ens.owner(name) == owner
    assert ens.address(name) == new_resolution
    ens.setup_name(name)
    assert ens.name(new_resolution) == name
    ens.setup_name(None, new_resolution)


def test_setup_name_default_to_owner(ens):
    name = 'reverse-defaults-to-owner.tester.eth'
    new_owner = ens.web3.eth.accounts[-1]
    ens.setup_owner(name, new_owner)
    assert not ens.name(new_owner)
    assert ens.owner(name) == new_owner
    assert not ens.address(name)
    ens.setup_name(name)
    assert ens.name(new_owner) == name


def test_setup_name_unowned_exception(ens):
    with pytest.raises(UnownedName):
        ens.setup_name('unowned-name.tester.eth')


def test_setup_name_unauthorized(ens, TEST_ADDRESS):
    with pytest.raises(UnauthorizedError):
        ens.setup_name('root-owned-tld', TEST_ADDRESS)


def test_setup_reverse_dict_unmodified(ens):
    # setup
    owner = ens.owner('tester.eth')
    eth = ens.web3.eth
    start_count = eth.getTransactionCount(owner)

    address = ens.web3.eth.accounts[3]
    transact = {}
    ens.setup_name('tester.eth', address, transact=transact)

    # even though a transaction was issued, the dict argument was not modified
    assert eth.getTransactionCount(owner) > start_count
    assert transact == {}

    # teardown
    ens.setup_name(None, address, transact=transact)



import pytest


@pytest.fixture(autouse=True)
def always_wait_for_mining_start(web3,
                                 wait_for_miner_start,
                                 skip_if_testrpc):
    skip_if_testrpc(web3)

    wait_for_miner_start(web3)

    assert web3.eth.mining
    assert web3.eth.hashrate
    assert web3.miner.hashrate



import pytest

from hypothesis import (
    given,
    strategies as st,
)


@st.composite
def dynamic_values(draw):
    matching_value = draw(st.text().filter(lambda x: x != ''))
    exclusions = (matching_value, '')
    non_matching_1 = draw(st.text().filter(lambda x: x not in exclusions))
    non_matching_2 = draw(st.text().filter(lambda x: x not in exclusions))
    non_matching_3 = draw(st.text().filter(lambda x: x not in exclusions))
    non_matching_4 = draw(st.text().filter(lambda x: x not in exclusions))
    return {
        "matching": matching_value,
        "non_matching": [
            non_matching_1,
            non_matching_2,
            non_matching_3,
            non_matching_4
        ]
    }


@st.composite
def fixed_values(draw):
    matching_values = draw(st.lists(elements=st.integers(min_value=0), min_size=4, max_size=4))
    non_matching_1 = draw(st.integers(min_value=0).filter(lambda x: x not in matching_values))
    non_matching_2 = draw(st.integers(min_value=0).filter(lambda x: x not in matching_values))
    non_matching_3 = draw(st.integers(min_value=0).filter(lambda x: x not in matching_values))
    non_matching_4 = draw(st.integers(min_value=0).filter(lambda x: x not in matching_values))
    return {
        "matching": matching_values,
        "non_matching": [
            non_matching_1,
            non_matching_2,
            non_matching_3,
            non_matching_4
        ]
    }


@st.composite
def array_values(draw):
    matching = draw(st.lists(elements=st.binary(min_size=2, max_size=2)))
    non_matching = draw(
        st.lists(elements=st.binary(min_size=2, max_size=2)).filter(lambda x: x != matching))
    return (matching, non_matching)


@pytest.mark.xfail(reason="The regex based data filters do not work with dynamic sized types")
@pytest.mark.parametrize('call_as_instance', (True, False))
@given(vals=dynamic_values())
def test_data_filters_with_dynamic_arguments(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter,
        vals):

    if call_as_instance:
        event_filter = create_filter(emitter, [
            'LogDynamicArgs', {
                'filter': {
                    'arg1': vals['matching']}}])
    else:
        event_filter = create_filter(Emitter, [
            'LogDynamicArgs', {
                'filter': {
                    'arg1': vals['matching']}}])

    txn_hashes = []
    txn_hashes.append(
        emitter.functions.logDynamicArgs(
            arg0=vals['matching'], arg1=vals['matching']).transact())
    txn_hashes.append(
        emitter.functions.logDynamicArgs(
            arg0=vals['non_matching'][0], arg1=vals['non_matching'][0]).transact())
    txn_hashes.append(
        emitter.functions.logDynamicArgs(
            arg0=vals['non_matching'][1], arg1=vals['non_matching'][1]).transact())
    txn_hashes.append(
        emitter.functions.logDynamicArgs(
            arg0=vals['non_matching'][2], arg1=vals['non_matching'][2]).transact())
    txn_hashes.append(
        emitter.functions.logDynamicArgs(
            arg0=vals['non_matching'][3], arg1=vals['non_matching'][3]).transact())

    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    log_entries = event_filter.get_new_entries()
    assert len(log_entries) == 1
    assert log_entries[0]['transactionHash'] == txn_hashes[0]


@pytest.mark.parametrize('call_as_instance', (True, False))
@given(vals=fixed_values())
def test_data_filters_with_fixed_arguments(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter,
        vals):

    if call_as_instance:
        event_filter = create_filter(emitter, [
            'LogQuadrupleArg', {
                'filter': {
                    'arg0': vals['matching'][0],
                    'arg1': vals['matching'][1],
                    'arg2': vals['matching'][2],
                    'arg3': vals['matching'][3]}}])
    else:
        event_filter = create_filter(Emitter, [
            'LogQuadrupleArg', {
                'filter': {
                    'arg0': vals['matching'][0],
                    'arg1': vals['matching'][1],
                    'arg2': vals['matching'][2],
                    'arg3': vals['matching'][3]}}])

    txn_hashes = []
    txn_hashes.append(emitter.functions.logQuadruple(
        which=5,
        arg0=vals['matching'][0],
        arg1=vals['matching'][1],
        arg2=vals['matching'][2],
        arg3=vals['matching'][3]).transact())
    txn_hashes.append(emitter.functions.logQuadruple(
        which=5,
        arg0=vals['non_matching'][0],
        arg1=vals['non_matching'][1],
        arg2=vals['non_matching'][2],
        arg3=vals['non_matching'][3]).transact())

    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    log_entries = event_filter.get_new_entries()
    assert len(log_entries) == 1
    assert log_entries[0]['transactionHash'] == txn_hashes[0]


@pytest.mark.xfail(reason="The regex based data filters do not work with dynamic sized types")
@pytest.mark.parametrize('call_as_instance', (True, False))
@given(vals=array_values())
def test_data_filters_with_list_arguments(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter,
        vals):

    matching, non_matching = vals

    if call_as_instance:
        event_filter = create_filter(emitter, [
            'LogListArgs', {
                'filter': {"arg1": matching}}])
    else:
        event_filter = create_filter(Emitter, [
            'LogListArgs', {
                'filter': {"arg1": matching}}])

    txn_hashes = []
    txn_hashes.append(emitter.functions.logListArgs(
        arg0=matching,
        arg1=matching).transact())
    txn_hashes.append(emitter.functions.logListArgs(
        arg0=non_matching,
        arg1=non_matching).transact())
    txn_hashes.append(emitter.functions.logListArgs(
        arg0=non_matching,
        arg1=matching).transact())
    txn_hashes.append(emitter.functions.logListArgs(
        arg0=matching,
        arg1=non_matching).transact())

    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    log_entries = event_filter.get_new_entries()
    assert len(log_entries) == 2
    assert log_entries[0]['transactionHash'] == txn_hashes[0]
    assert log_entries[1]['transactionHash'] == txn_hashes[3]




import pytest

from web3.utils.formatters import (
    apply_formatters_to_dict,
    map_collection,
    recursive_map,
)


def square_int(x):
    if isinstance(x, int):
        return x * x
    else:
        return x


@pytest.mark.parametrize('non_collection', [1, 'abc', u'def', True, None])
def test_map_collection_on_non_collection(non_collection):
    assert map_collection(lambda x: x + 2, non_collection) == non_collection


@pytest.mark.parametrize('coll', [set, list, tuple])
def test_collection_apply(coll):
    vals = coll([1, 2])
    assert map_collection(lambda x: x + 2, vals) == coll([3, 4])


def test_collection_apply_to_mapping():
    vals = {'a': 1, 'b': 2}
    assert map_collection(lambda x: x + 2, vals) == {'a': 3, 'b': 4}


def test_recursive_collection_apply():
    assert recursive_map(square_int, [[3]]) == [[9]]


def test_recursive_collection_cycle():
    data = [3]
    data.append(data)
    with pytest.raises(ValueError):
        recursive_map(square_int, data)


def test_format_dict_error():
    with pytest.raises(ValueError) as exc_info:
        apply_formatters_to_dict(
            {'myfield': int},
            {'myfield': 'a'},
        )
    assert 'myfield' in str(exc_info.value)



import os
import pytest
import tempfile

from tests.integration.parity.utils import (
    wait_for_socket,
)
from web3 import Web3
from web3.utils.module_testing import (
    NetModuleTest,
    VersionModuleTest,
)

from .common import (
    ParityEthModuleTest,
    ParityPersonalModuleTest,
    ParityTraceModuleTest,
    ParityWeb3ModuleTest,
)


@pytest.fixture(scope='module')
def ipc_path(datadir):
    ipc_dir_path = tempfile.mkdtemp()
    _ipc_path = os.path.join(ipc_dir_path, 'jsonrpc.ipc')
    yield _ipc_path

    if os.path.exists(_ipc_path):
        os.remove(_ipc_path)


@pytest.fixture(scope="module")
def parity_command_arguments(
    parity_import_blocks_process,
    parity_binary,
    ipc_path,
    datadir,
    passwordfile,
    author
):
    return (
        parity_binary,
        '--chain', os.path.join(datadir, 'chain_config.json'),
        '--ipc-path', ipc_path,
        '--base-path', datadir,
        '--unlock', author,
        '--password', passwordfile,
        '--no-jsonrpc',
        '--no-ws',
    )


@pytest.fixture(scope="module")
def parity_import_blocks_command(parity_binary, ipc_path, datadir, passwordfile):
    return (
        parity_binary,
        'import', os.path.join(datadir, 'blocks_export.rlp'),
        '--chain', os.path.join(datadir, 'chain_config.json'),
        '--ipc-path', ipc_path,
        '--base-path', datadir,
        '--password', passwordfile,
        '--no-jsonrpc',
        '--no-ws',
        '--tracing', 'on',
    )


@pytest.fixture(scope="module")  # noqa: F811
def web3(parity_process, ipc_path):
    wait_for_socket(ipc_path)
    _web3 = Web3(Web3.IPCProvider(ipc_path))
    return _web3


class TestParityWeb3ModuleTest(ParityWeb3ModuleTest):
    pass


class TestParityEthModuleTest(ParityEthModuleTest):
    pass


class TestParityVersionModule(VersionModuleTest):
    pass


class TestParityNetModule(NetModuleTest):
    pass


class TestParityPersonalModuleTest(ParityPersonalModuleTest):
    pass


class TestParityTraceModuleTest(ParityTraceModuleTest):
    pass



import pytest

from web3.utils.abi import (
    is_encodable,
)


@pytest.mark.parametrize(
    'value,_type,expected',
    (
        # bytes
        ('12', 'bytes2', True),  # undersize OK
        ('0x12', 'bytes2', True),  # with or without 0x OK
        ('0123', 'bytes2', True),  # exact size OK
        (b'\x12', 'bytes2', True),  # as bytes value undersize OK
        (b'\x01\x23', 'bytes2', True),  # as bytes value exact size OK
        (b'\x01\x23', 'bytes1', False),  # no oversize bytes
        ('0123', 'bytes1', False),  # no oversize hex strings
        ('1', 'bytes2', False),  # no odd length
        ('0x1', 'bytes2', False),  # no odd length
        (True, 'bytes32', False),  # no wrong types
        (0, 'bytes32', False),  # no wrong types
        # int
        (-1 * 2**255 - 1, 'int256', False),
        (-1 * 2**255, 'int256', True),
        (-1, 'int256', True),
        (0, 'int256', True),
        (1, 'int256', True),
        (2**255 - 1, 'int256', True),
        (2**255, 'int256', False),
        ('abc', 'int256', False),
        (True, 'int256', False),
        # uint
        (-1, 'uint256', False),
        (0, 'uint256', True),
        (1, 'uint256', True),
        (2**256 - 1, 'uint256', True),
        (2**256, 'uint256', False),
        ('abc', 'uint256', False),
        (True, 'uint256', False),
        # function
        (0, 'function', False),
        (b'\0' * 24, 'function', True),
        (b'\0' * 25, 'function', False),
        (True, 'function', False),
        (False, 'function', False),
        # address
        ('0x' + '00' * 20, 'address', True),
        ('0x' + '00' * 32, 'address', False),
        (None, 'address', False),
        ('dennisthepeasant.eth', 'address', True),  # passes because eth_utils converts to bytes :/
        ('autonomouscollective.eth', 'address', True),
        ('all-TLDs-valid-now.test', 'address', True),
        ('-rejects-invalid-names.test', 'address', False),
        ('ff', 'address', True),  # this could theoretically be a top-level domain (TLD)
        ('0xname.eth', 'address', True),  # 0x in name is fine, if it is not a TLD
        ('0xff', 'address', False),  # but any valid hex starting with 0x should be rejected
        # string
        ('', 'string', True),
        ('anything', 'string', True),
        (b'', 'string', True),
        (b'anything', 'string', True),
        (b'\x80', 'string', False),  # bytes that cannot be decoded with utf-8 are invalid
    ),
)
def test_is_encodable(value, _type, expected):
    actual = is_encodable(_type, value)
    assert actual is expected



import pytest

from web3.exceptions import (
    ValidationError,
)
from web3.middleware import (
    construct_fixture_middleware,
    geth_poa_middleware,
)


# In the spec, a block with extra data longer than 32 bytes is invalid
def test_long_extra_data(web3):
    return_block_with_long_extra_data = construct_fixture_middleware({
        'eth_getBlockByNumber': {'extraData': '0x' + 'ff' * 33},
    })
    web3.middleware_stack.inject(return_block_with_long_extra_data, layer=0)
    with pytest.raises(ValidationError):
        web3.eth.getBlock('latest')


def test_full_extra_data(web3):
    return_block_with_long_extra_data = construct_fixture_middleware({
        'eth_getBlockByNumber': {'extraData': '0x' + 'ff' * 32},
    })
    web3.middleware_stack.inject(return_block_with_long_extra_data, layer=0)
    block = web3.eth.getBlock('latest')
    assert block.extraData == b'\xff' * 32


def test_geth_proof_of_authority(web3):
    return_block_with_long_extra_data = construct_fixture_middleware({
        'eth_getBlockByNumber': {'extraData': '0x' + 'ff' * 33},
    })
    web3.middleware_stack.inject(geth_poa_middleware, layer=0)
    web3.middleware_stack.inject(return_block_with_long_extra_data, layer=0)
    block = web3.eth.getBlock('latest')
    assert 'extraData' not in block
    assert block.proofOfAuthorityData == b'\xff' * 33



from binascii import (
    unhexlify,
)
import json
import pytest

ABI_A = json.loads('[{"constant":false,"inputs":[],"name":"noargfunc","outputs":[],"type":"function"}]')  # noqa: E501
ABI_B = json.loads('[{"constant":false,"inputs":[{"name":"uintarg","type":"uint256"}],"name":"uintfunc","outputs":[],"type":"function"}]')  # noqa: E501
ABI_C = json.loads('[{"constant":false,"inputs":[],"name":"namesakefunc","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"bytesarg","type":"bytes32"}],"name":"namesakefunc","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"uintarg","type":"uint256"}],"name":"namesakefunc","outputs":[],"type":"function"}]')  # noqa: E501
ABI_D = json.loads('[{ "constant": false, "inputs": [ { "name": "b", "type": "bytes32[]" } ], "name": "byte_array", "outputs": [], "payable": false, "type": "function" }]')  # noqa: E501
ABI_BYTES = json.loads('[{"constant":false,"inputs":[{"name":"bytesarg","type":"bytes"}],"name":"bytesfunc","outputs":[],"type":"function"}]')  # noqa: E501
ABI_STRING = json.loads('[{"constant":false,"inputs":[{"name":"stringarg","type":"string"}],"name":"stringfunc","outputs":[],"type":"function"}]')  # noqa: E501
ABI_ADDRESS = json.loads('[{"constant":false,"inputs":[{"name":"addressarg","type":"address"}],"name":"addressfunc","outputs":[],"type":"function"}]')  # noqa: E501
a32bytes = b'a'.ljust(32, b'\x00')


@pytest.mark.parametrize(
    'abi,data,method,expected',
    (
        (
            ABI_A,
            '0xc4c1a40b',
            'noargfunc',
            {},
        ),
        (
            ABI_B,
            '0xcc6820de0000000000000000000000000000000000000000000000000000000000000001',
            'uintfunc',
            {'uintarg': 1},
        ),
        (
            ABI_C,
            '0x22d86fa3',
            'namesakefunc',
            {},
        ),
        (
            ABI_C,
            '0x40c05b2f0000000000000000000000000000000000000000000000000000000000000001',
            'namesakefunc',
            {'uintarg': 1},
        ),
        (
            ABI_C,
            '0xf931d77c6100000000000000000000000000000000000000000000000000000000000000',
            'namesakefunc',
            {'bytesarg': a32bytes},
        ),
        (
            ABI_BYTES,
            '0xb606a9f6000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000016100000000000000000000000000000000000000000000000000000000000000',  # noqa: E501
            'bytesfunc',
            {'bytesarg': b'a'},
        ),
        (
            ABI_STRING,
            '0x33b4005f000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000016100000000000000000000000000000000000000000000000000000000000000',  # noqa: E501
            'stringfunc',
            {'stringarg': 'a'},
        ),
        (
            ABI_ADDRESS,
            '0x4767be6c000000000000000000000000ffffffffffffffffffffffffffffffffffffffff',
            'addressfunc',
            {'addressarg': '0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF'},
        ),
    ),
)
def test_contract_abi_decoding(web3, abi, data, method, expected):
    contract = web3.eth.contract(abi=abi)
    func, params = contract.decode_function_input(data)
    assert func.fn_name == method
    assert params == expected

    reinvoke_func = contract.functions[func.fn_name](**params)
    rebuild_txn = reinvoke_func.buildTransaction({'gas': 0, 'nonce': 0, 'to': '\x00' * 20})
    assert rebuild_txn['data'] == data


@pytest.mark.parametrize(
    'abi,method,expected,data',
    (
        (
            ABI_D,
            'byte_array',
            {
                'b': [
                    unhexlify('5595c210956e7721f9b692e702708556aa9aabb14ea163e96afa56ffbe9fa809'),
                    unhexlify('6f8d2fa18448afbfe4f82143c384484ad09a0271f3a3c0eb9f629e703f883125'),
                ],
            },
            '0xf166d6f8000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000025595c210956e7721f9b692e702708556aa9aabb14ea163e96afa56ffbe9fa8096f8d2fa18448afbfe4f82143c384484ad09a0271f3a3c0eb9f629e703f883125',  # noqa: E501
        ),
    ),
)
def test_contract_abi_encoding_kwargs(web3, abi, method, expected, data):
    contract = web3.eth.contract(abi=abi)
    func, params = contract.decode_function_input(data)
    assert func.fn_name == method
    assert params == expected

    reinvoke_func = contract.functions[func.fn_name](**params)
    rebuild_txn = reinvoke_func.buildTransaction({'gas': 0, 'nonce': 0, 'to': '\x00' * 20})
    assert rebuild_txn['data'] == data



import pytest

from web3 import Web3

from .common import (
    GoEthereumEthModuleTest,
    GoEthereumNetModuleTest,
    GoEthereumPersonalModuleTest,
    GoEthereumTest,
    GoEthereumVersionModuleTest,
    get_open_port,
)
from .utils import (
    wait_for_http,
)


@pytest.fixture(scope="module")
def rpc_port():
    return get_open_port()


@pytest.fixture(scope="module")
def endpoint_uri(rpc_port):
    return 'http://localhost:{0}'.format(rpc_port)


@pytest.fixture(scope='module')
def geth_command_arguments(geth_binary, datadir, rpc_port):
    return (
        geth_binary,
        '--datadir', str(datadir),
        '--nodiscover',
        '--fakepow',
        '--rpc',
        '--rpcport', rpc_port,
        '--rpcapi', 'db,eth,net,web3,personal,web3',
        '--ipcdisable',
    )


@pytest.fixture(scope="module")
def web3(geth_process, endpoint_uri):
    wait_for_http(endpoint_uri)
    _web3 = Web3(Web3.HTTPProvider(endpoint_uri))
    return _web3


class TestGoEthereumTest(GoEthereumTest):
    pass


class TestGoEthereumEthModuleTest(GoEthereumEthModuleTest):
    pass


class TestGoEthereumVersionModuleTest(GoEthereumVersionModuleTest):
    pass


class TestGoEthereumNetModuleTest(GoEthereumNetModuleTest):
    pass


class TestGoEthereumPersonalModuleTest(GoEthereumPersonalModuleTest):
    pass



import logging
import os

from web3 import (
    HTTPProvider,
    Web3,
)

INFURA_MAINNET_BASE_URL = 'https://mainnet.infura.io'


def load_infura_url():
    key = os.environ.get('INFURA_API_KEY', '')
    if key == '':
        logging.getLogger('web3.auto.infura').warning(
            "No Infura API Key found. Add environment variable INFURA_API_KEY to ensure continued "
            "API access. New keys are available at https://infura.io/signup"
        )
        return INFURA_MAINNET_BASE_URL
    else:
        return "%s/%s" % (INFURA_MAINNET_BASE_URL, key)


w3 = Web3(HTTPProvider(load_infura_url()))



def test_testing_mine_single_block(web3):
    web3.testing.mine()

    before_mining_block = web3.eth.getBlock("latest")

    web3.testing.mine()

    after_mining_block = web3.eth.getBlock("latest")

    assert after_mining_block['number'] - before_mining_block['number'] == 1


def test_testing_mine_multiple_blocks(web3):
    web3.testing.mine()

    before_mining_block = web3.eth.getBlock("latest")

    web3.testing.mine(5)

    after_mining_block = web3.eth.getBlock("latest")

    assert after_mining_block['number'] - before_mining_block['number'] == 5



import pytest

from web3.utils.abi import (
    filter_by_argument_name,
)

ABI = [
    {
        "constant": False,
        "inputs": [],
        "name": "func_1",
        "outputs": [],
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"name": "a", "type": "uint256"},
        ],
        "name": "func_2",
        "outputs": [],
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"name": "a", "type": "uint256"},
            {"name": "b", "type": "uint256"},
        ],
        "name": "func_3",
        "outputs": [],
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"name": "a", "type": "uint256"},
            {"name": "b", "type": "uint256"},
            {"name": "c", "type": "uint256"},
        ],
        "name": "func_4",
        "outputs": [],
        "type": "function",
    },
]


@pytest.mark.parametrize(
    'argument_names,expected',
    (
        ([], ['func_1', 'func_2', 'func_3', 'func_4']),
        (['a'], ['func_2', 'func_3', 'func_4']),
        (['a', 'c'], ['func_4']),
        (['c'], ['func_4']),
        (['b'], ['func_3', 'func_4']),
    )
)
def test_filter_by_arguments_1(argument_names, expected):
    actual_matches = filter_by_argument_name(argument_names, ABI)
    function_names = [match['name'] for match in actual_matches]
    assert set(function_names) == set(expected)




import pytest

from ens import (
    InvalidName,
)

# test content inspired by https://github.com/jcranmer/idna-uts46/blob/9e2ff191a8887c872ad172f2d72f8d4bf353aef8/test/test-uts46.js  # noqa: E501


def test_nameprep_basic_unicode(ens):
    assert ens.nameprep("öbb.at") == "öbb.at"
    assert ens.nameprep("Öbb.at") == "öbb.at"
    assert ens.nameprep("O\u0308bb.at") == "öbb.at"
    assert ens.nameprep("xn--bb-eka.at") == "öbb.at"
    assert ens.nameprep("faß.de") == "faß.de"
    assert ens.nameprep("fass.de") == "fass.de"
    assert ens.nameprep("xn--fa-hia.de") == "faß.de"


def test_nameprep_std3_rules(ens):
    with pytest.raises(InvalidName):
        ens.nameprep("not=std3")



import pytest

from eth_utils import (
    to_bytes,
)

from web3.utils.toolz import (
    identity,
)


@pytest.fixture(scope="module", params=[lambda x: to_bytes(hexstr=x), identity])
def address_conversion_func(request):
    return request.param



import pytest

from web3.utils.abi import (
    is_address_type,
    is_array_type,
    is_bool_type,
    is_bytes_type,
    is_int_type,
    is_recognized_type,
    is_string_type,
    is_uint_type,
)


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('bool', True),
        ('uint', False),
        ('uint256', True),
        ('uint8', True),
        ('uint7', False),
        ('int', False),
        ('int256', True),
        ('int8', True),
        ('int7', False),
        ('byte', False),
        ('bytes1', True),
        ('bytes7', True),
        ('bytes32', True),
        ('bytes32.byte', True),
        ('bytes', True),
        ('string', True),
        ('address', True),
        ('uint256[]', True),
        ('uint256[][]', True),
        ('uint256[][][]', True),
        ('uint256[1][][3]', True),
        ('uint256[1][20000][3]', True),
        ('uint256[][20000][]', True),
        ('bytes20[]', True),
        ('bytes20[][]', True),
        ('bytes20[][][]', True),
        ('bytes20[1][][3]', True),
        ('bytes20[1][20000][3]', True),
        ('bytes20[][20000][]', True),
        ('address[]', True),
        ('address[][]', True),
        ('address[][][]', True),
        ('address[1][][3]', True),
        ('address[1][20000][3]', True),
        ('address[][20000][]', True),
        ('SomeEnum.SomeValue', False),
    ),
)
def test_is_recognized_type(abi_type, should_match):
    is_match = is_recognized_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('bool', True),
        ('uint', False),
        ('sbool', False),
        ('bools', False),
    ),
)
def test_is_bool_type(abi_type, should_match):
    is_match = is_bool_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('uint', False),
        ('uint32', True),
        ('uint255', False),
        ('uint256', True),
        ('uint0', False),
        ('int', False),
        ('int16', False),
        ('suint', False),
        ('uints', False),
    ),
)
def test_is_uint_type(abi_type, should_match):
    is_match = is_uint_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('int', False),
        ('int32', True),
        ('int255', False),
        ('int256', True),
        ('int0', False),
        ('uint', False),
        ('uint16', False),
        ('sint', False),
        ('ints', False),
    ),
)
def test_is_int_type(abi_type, should_match):
    is_match = is_int_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('address', True),
        ('uint', False),
        ('saddress', False),
        ('addresss', False),
    ),
)
def test_is_address_type(abi_type, should_match):
    is_match = is_address_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('bytes', True),
        ('bytes4', True),
        ('bytes32.byte', True),
        ('sbyte', False),
        ('bytess', False),
        ('byte', False),
        ('byte0', False),
        ('bytes33', False),
        ('bytes32..byte', False),
        ('bytes32mbyte', False),
        ('bytes32byte', False),

    ),
)
def test_is_bytes_type(abi_type, should_match):
    is_match = is_bytes_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('string', True),
        ('uint', False),
        ('sstring', False),
        ('strings', False),
    ),
)
def test_is_string_type(abi_type, should_match):
    is_match = is_string_type(abi_type)
    assert is_match is should_match


@pytest.mark.parametrize(
    'abi_type,should_match',
    (
        ('bool[]', True),
        ('uint[]', True),
        ('uint[][]', True),
        ('uint[5][]', True),
        ('uint[][5]', True),
        ('int[]', True),
        ('string[]', True),
        ('address[]', True),
        ('bytes[]', True),
        ('string', False),
        ('bytes', False),
        ('uint[', False),
        ('uint]', False),
    ),
)
def test_is_array_type(abi_type, should_match):
    is_match = is_array_type(abi_type)
    assert is_match is should_match



import pytest

from hypothesis import (
    given,
    strategies as st,
)

from web3.exceptions import (
    InsufficientData,
)
from web3.utils.math import (
    percentile,
)

values = range(100)


@pytest.mark.parametrize(
    "p,expected",
    (
        (0, 0),
        (1, 0),
        (2, 1),
        (3, 2),
        (4, 3),
        (50, 49),
        (50.5, 49.5),
        (100, 99)
    )
)
def test_percentiles_out_of_one_hundred(p, expected):
    assert percentile(values, p) == expected


def test_percentiles_with_no_values():
    with pytest.raises(InsufficientData):
        percentile([], 1)


def test_percentiles_with_out_of_bounds_fractions():
    assert 1 == percentile([1, 2, 3, 4], percentile=10)
    assert 1 == percentile([1, 2, 3, 4], percentile=15)
    assert 1 == percentile([1, 2, 3, 4], percentile=20)
    assert 1 == percentile([1, 2, 3, 4], percentile=25)
    assert 1 < percentile([1, 2, 3, 4], percentile=30)


@given(
    values=st.lists(elements=st.integers(), min_size=1, max_size=200),
    p=st.integers(max_value=100, min_value=0))
def test_fuzz_test_percentiles(values, p):
    if not values:
        with pytest.raises(ValueError):
            percentile(values, p)
    else:
        percentile(values, p)



from decimal import (
    Decimal,
    getcontext,
)
from distutils.version import (
    LooseVersion,
)
import json
import pytest

import eth_abi
from hexbytes import (
    HexBytes,
)

from web3.exceptions import (
    BadFunctionCallOutput,
    BlockNumberOutofRange,
    InvalidAddress,
    MismatchedABI,
    NoABIFunctionsFound,
    ValidationError,
)
from web3.utils.ens import (
    contract_ens_addresses,
)
from web3.utils.toolz import (
    identity,
)

# Ignore warning in pyethereum 1.6 - will go away with the upgrade
pytestmark = pytest.mark.filterwarnings("ignore:implicit cast from 'char *'")


def deploy(web3, Contract, apply_func=identity, args=None):
    args = args or []
    deploy_txn = Contract.constructor(*args).transact()
    deploy_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert deploy_receipt is not None
    address = apply_func(deploy_receipt['contractAddress'])
    contract = Contract(address=address)
    assert contract.address == address
    assert len(web3.eth.getCode(contract.address)) > 0
    return contract


@pytest.fixture()
def address_reflector_contract(web3, AddressReflectorContract, address_conversion_func):
    return deploy(web3, AddressReflectorContract, address_conversion_func)


@pytest.fixture()
def math_contract(web3, MathContract, address_conversion_func):
    return deploy(web3, MathContract, address_conversion_func)


@pytest.fixture()
def string_contract(web3, StringContract, address_conversion_func):
    return deploy(web3, StringContract, address_conversion_func, args=["Caqalai"])


@pytest.fixture()
def arrays_contract(web3, ArraysContract, address_conversion_func):
    # bytes_32 = [keccak('0'), keccak('1')]
    bytes32_array = [
        b'\x04HR\xb2\xa6p\xad\xe5@~x\xfb(c\xc5\x1d\xe9\xfc\xb9eB\xa0q\x86\xfe:\xed\xa6\xbb\x8a\x11m',  # noqa: E501
        b'\xc8\x9e\xfd\xaaT\xc0\xf2\x0cz\xdfa(\x82\xdf\tP\xf5\xa9Qc~\x03\x07\xcd\xcbLg/)\x8b\x8b\xc6',  # noqa: E501
    ]
    byte_arr = [b'\xff', b'\xff', b'\xff', b'\xff']
    return deploy(web3, ArraysContract, address_conversion_func, args=[bytes32_array, byte_arr])


@pytest.fixture()
def address_contract(web3, WithConstructorAddressArgumentsContract, address_conversion_func):
    return deploy(
        web3,
        WithConstructorAddressArgumentsContract,
        address_conversion_func,
        args=["0xd3CdA913deB6f67967B99D67aCDFa1712C293601"]
    )


@pytest.fixture(params=[b'\x04\x06', '0x0406', '0406'])
def bytes_contract(web3, BytesContract, request, address_conversion_func):
    return deploy(web3, BytesContract, address_conversion_func, args=[request.param])


@pytest.fixture()
def fixed_reflection_contract(web3, FixedReflectionContract, address_conversion_func):
    return deploy(web3, FixedReflectionContract, address_conversion_func)


@pytest.fixture()
def call_transaction():
    return {
        'data': '0x61bc221a',
        'to': '0xc305c901078781C232A2a521C2aF7980f8385ee9'
    }


@pytest.fixture(params=[
    '0x0406040604060406040604060406040604060406040604060406040604060406',
    '0406040604060406040604060406040604060406040604060406040604060406',
    HexBytes('0406040604060406040604060406040604060406040604060406040604060406'),
])
def bytes32_contract(web3, Bytes32Contract, request, address_conversion_func):
    return deploy(web3, Bytes32Contract, address_conversion_func, args=[request.param])


@pytest.fixture()
def undeployed_math_contract(web3, MathContract, address_conversion_func):
    empty_address = address_conversion_func("0x000000000000000000000000000000000000dEaD")
    _undeployed_math_contract = MathContract(address=empty_address)
    return _undeployed_math_contract


@pytest.fixture()
def mismatched_math_contract(web3, StringContract, MathContract, address_conversion_func):
    deploy_txn = StringContract.constructor("Caqalai").transact()
    deploy_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert deploy_receipt is not None
    address = address_conversion_func(deploy_receipt['contractAddress'])
    _mismatched_math_contract = MathContract(address=address)
    return _mismatched_math_contract


@pytest.fixture()
def fallback_function_contract(web3, FallballFunctionContract, address_conversion_func):
    return deploy(web3, FallballFunctionContract, address_conversion_func)


def test_invalid_address_in_deploy_arg(web3, WithConstructorAddressArgumentsContract):
    with pytest.raises(InvalidAddress):
        WithConstructorAddressArgumentsContract.constructor(
            "0xd3cda913deb6f67967b99d67acdfa1712c293601",
        ).transact()


def test_call_with_no_arguments(math_contract, call):
    result = call(contract=math_contract,
                  contract_function='return13')
    assert result == 13


def test_call_with_one_argument(math_contract, call):
    result = call(contract=math_contract,
                  contract_function='multiply7',
                  func_args=[3])
    assert result == 21


@pytest.mark.parametrize(
    'call_args,call_kwargs',
    (
        ((9, 7), {}),
        ((9,), {'b': 7}),
        (tuple(), {'a': 9, 'b': 7}),
    ),
)
def test_call_with_multiple_arguments(math_contract, call, call_args, call_kwargs):
    result = call(contract=math_contract,
                  contract_function='add',
                  func_args=call_args,
                  func_kwargs=call_kwargs)
    assert result == 16


@pytest.mark.parametrize(
    'call_args,call_kwargs',
    (
        ((9, 7), {}),
        ((9,), {'b': 7}),
        (tuple(), {'a': 9, 'b': 7}),
    ),
)
def test_saved_method_call_with_multiple_arguments(math_contract, call_args, call_kwargs):
    math_contract_add = math_contract.functions.add(*call_args, **call_kwargs)
    result = math_contract_add.call()
    assert result == 16


def test_call_get_string_value(string_contract, call):
    result = call(contract=string_contract,
                  contract_function='getValue')
    # eth_abi.decode_api() does not assume implicit utf-8
    # encoding of string return values. Thus, we need to decode
    # ourselves for fair comparison.
    assert result == "Caqalai"


@pytest.mark.skipif(
    LooseVersion(eth_abi.__version__) >= LooseVersion("2"),
    reason="eth-abi >=2 does utf-8 string decoding")
def test_call_read_string_variable(string_contract, call):
    result = call(contract=string_contract,
                  contract_function='constValue')
    assert result == b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6a\x6b\x6c\x6d\x6e\x6f\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7a\x7b\x7c\x7d\x7e\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff".decode(errors='backslashreplace')  # noqa: E501


@pytest.mark.skipif(
    LooseVersion(eth_abi.__version__) < LooseVersion("2"),
    reason="eth-abi does not raise exception on undecodable bytestrings")
def test_call_on_undecodable_string(string_contract, call):
    with pytest.raises(BadFunctionCallOutput):
        call(
            contract=string_contract,
            contract_function='constValue')


def test_call_get_bytes32_array(arrays_contract, call):
    result = call(contract=arrays_contract,
                  contract_function='getBytes32Value')
    # expected_bytes32_array = [keccak('0'), keccak('1')]
    expected_bytes32_array = [
        b'\x04HR\xb2\xa6p\xad\xe5@~x\xfb(c\xc5\x1d\xe9\xfc\xb9eB\xa0q\x86\xfe:\xed\xa6\xbb\x8a\x11m',  # noqa: E501
        b'\xc8\x9e\xfd\xaaT\xc0\xf2\x0cz\xdfa(\x82\xdf\tP\xf5\xa9Qc~\x03\x07\xcd\xcbLg/)\x8b\x8b\xc6',  # noqa: E501
    ]
    assert result == expected_bytes32_array


def test_call_get_bytes32_const_array(arrays_contract, call):
    result = call(contract=arrays_contract,
                  contract_function='getBytes32ConstValue')
    # expected_bytes32_array = [keccak('A'), keccak('B')]
    expected_bytes32_array = [
        b'\x03x?\xac.\xfe\xd8\xfb\xc9\xadD>Y.\xe3\x0ea\xd6_G\x11@\xc1\x0c\xa1U\xe97\xb45\xb7`',
        b'\x1fg[\xff\x07Q_]\xf9g7\x19N\xa9E\xc3lA\xe7\xb4\xfc\xef0{|\xd4\xd0\xe6\x02\xa6\x91\x11',
    ]
    assert result == expected_bytes32_array


def test_call_get_byte_array(arrays_contract, call):
    result = call(contract=arrays_contract,
                  contract_function='getByteValue')
    expected_byte_arr = [b'\xff', b'\xff', b'\xff', b'\xff']
    assert result == expected_byte_arr


def test_call_get_byte_const_array(arrays_contract, call):
    result = call(contract=arrays_contract,
                  contract_function='getByteConstValue')
    expected_byte_arr = [b'\x00', b'\x01']
    assert result == expected_byte_arr


def test_call_read_address_variable(address_contract, call):
    result = call(contract=address_contract,
                  contract_function='testAddr')
    assert result == "0xd3CdA913deB6f67967B99D67aCDFa1712C293601"


def test_init_with_ens_name_arg(web3, WithConstructorAddressArgumentsContract, call):
    with contract_ens_addresses(
        WithConstructorAddressArgumentsContract,
        [("arg-name.eth", "0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413")],
    ):
        address_contract = deploy(web3, WithConstructorAddressArgumentsContract, args=[
            "arg-name.eth",
        ])

    result = call(contract=address_contract,
                  contract_function='testAddr')
    assert result == "0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413"


def test_call_read_bytes_variable(bytes_contract, call):
    result = call(contract=bytes_contract, contract_function='constValue')
    assert result == b"\x01\x23"


def test_call_get_bytes_value(bytes_contract, call):
    result = call(contract=bytes_contract, contract_function='getValue')
    assert result == b'\x04\x06'


def test_call_read_bytes32_variable(bytes32_contract, call):
    result = call(contract=bytes32_contract, contract_function='constValue')
    assert result == b"\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23\x01\x23"  # noqa


def test_call_get_bytes32_value(bytes32_contract, call):
    result = call(contract=bytes32_contract, contract_function='getValue')
    assert result == b'\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06\x04\x06'  # noqa


@pytest.mark.parametrize(
    'value, expected',
    [
        (
            '0x' + '11' * 20,
            '0x' + '11' * 20,
        ),
        (
            '0xbb9bc244d798123fde783fcc1c72d3bb8c189413',
            InvalidAddress,
        ),
        (
            '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413',
            '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413',
        ),
    ]
)
def test_call_address_reflector_with_address(address_reflector_contract, value, expected, call):
    if not isinstance(expected, str):
        with pytest.raises(expected):
            call(contract=address_reflector_contract,
                 contract_function='reflect',
                 func_args=[value])
    else:
        assert call(contract=address_reflector_contract,
                    contract_function='reflect',
                    func_args=[value]) == expected


@pytest.mark.parametrize(
    'value, expected',
    [
        (
            ['0x' + '11' * 20, '0x' + '22' * 20],
            ['0x' + '11' * 20, '0x' + '22' * 20],
        ),
        (
            ['0x' + '11' * 20, '0x' + 'aa' * 20],
            InvalidAddress
        ),
        (
            [
                '0xFeC2079e80465cc8C687fFF9EE6386ca447aFec4',
                '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413',
            ],
            [
                '0xFeC2079e80465cc8C687fFF9EE6386ca447aFec4',
                '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413',
            ],
        ),
    ]
)
def test_call_address_list_reflector_with_address(address_reflector_contract,
                                                  value,
                                                  expected,
                                                  call):
    if not isinstance(expected, list):
        with pytest.raises(expected):
            call(contract=address_reflector_contract,
                 contract_function='reflect',
                 func_args=[value])
    else:
        assert call(contract=address_reflector_contract,
                    contract_function='reflect',
                    func_args=[value]) == expected


def test_call_address_reflector_single_name(address_reflector_contract, call):
    with contract_ens_addresses(
        address_reflector_contract,
        [("dennisthepeasant.eth", "0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413")],
    ):
        result = call(contract=address_reflector_contract,
                      contract_function='reflect',
                      func_args=['dennisthepeasant.eth'])
        assert result == '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413'


def test_call_address_reflector_name_array(address_reflector_contract, call):
    names = [
        'autonomouscollective.eth',
        'wedonthavealord.eth',
    ]
    addresses = [
        '0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413',
        '0xFeC2079e80465cc8C687fFF9EE6386ca447aFec4',
    ]

    with contract_ens_addresses(address_reflector_contract, zip(names, addresses)):
        result = call(contract=address_reflector_contract,
                      contract_function='reflect',
                      func_args=[names])

    assert addresses == result


def test_call_reject_invalid_ens_name(address_reflector_contract, call):
    with contract_ens_addresses(address_reflector_contract, []):
        with pytest.raises(ValueError):
            call(contract=address_reflector_contract,
                 contract_function='reflect',
                 func_args=['type0.eth'])


def test_call_missing_function(mismatched_math_contract, call):
    expected_missing_function_error_message = "Could not decode contract function call"
    with pytest.raises(BadFunctionCallOutput) as exception_info:
        call(contract=mismatched_math_contract, contract_function='return13')
    assert expected_missing_function_error_message in str(exception_info.value)


def test_call_undeployed_contract(undeployed_math_contract, call):
    expected_undeployed_call_error_message = "Could not transact with/call contract function"
    with pytest.raises(BadFunctionCallOutput) as exception_info:
        call(contract=undeployed_math_contract, contract_function='return13')
    assert expected_undeployed_call_error_message in str(exception_info.value)


def test_call_fallback_function(fallback_function_contract):
    result = fallback_function_contract.fallback.call()
    assert result == []


def test_throws_error_if_block_out_of_range(web3, math_contract):
    web3.providers[0].make_request(method='evm_mine', params=[20])
    with pytest.raises(BlockNumberOutofRange):
        math_contract.functions.counter().call(block_identifier=-50)


def test_accepts_latest_block(web3, math_contract):
    web3.providers[0].make_request(method='evm_mine', params=[5])
    math_contract.functions.increment().transact()

    late = math_contract.functions.counter().call(block_identifier='latest')
    pend = math_contract.functions.counter().call(block_identifier='pending')

    assert late == 1
    assert pend == 1


def test_accepts_block_hash_as_identifier(web3, math_contract):
    blocks = web3.providers[0].make_request(method='evm_mine', params=[5])
    math_contract.functions.increment().transact()
    more_blocks = web3.providers[0].make_request(method='evm_mine', params=[5])

    old = math_contract.functions.counter().call(block_identifier=blocks['result'][2])
    new = math_contract.functions.counter().call(block_identifier=more_blocks['result'][2])

    assert old == 0
    assert new == 1


def test_neg_block_indexes_from_the_end(web3, math_contract):
    web3.providers[0].make_request(method='evm_mine', params=[5])
    math_contract.functions.increment().transact()
    math_contract.functions.increment().transact()
    web3.providers[0].make_request(method='evm_mine', params=[5])

    output1 = math_contract.functions.counter().call(block_identifier=-7)
    output2 = math_contract.functions.counter().call(block_identifier=-6)

    assert output1 == 1
    assert output2 == 2


def test_returns_data_from_specified_block(web3, math_contract):
    start_num = web3.eth.getBlock('latest').number
    web3.providers[0].make_request(method='evm_mine', params=[5])
    math_contract.functions.increment().transact()
    math_contract.functions.increment().transact()

    output1 = math_contract.functions.counter().call(block_identifier=start_num + 6)
    output2 = math_contract.functions.counter().call(block_identifier=start_num + 7)

    assert output1 == 1
    assert output2 == 2


message_regex = (
    r"\nCould not identify the intended function with name `.*`, "
    r"positional argument\(s\) of type `.*` and "
    r"keyword argument\(s\) of type `.*`."
    r"\nFound .* function\(s\) with the name `.*`: .*"
)
diagnosis_arg_regex = (
    r"\nFunction invocation failed due to improper number of arguments."
)
diagnosis_encoding_regex = (
    r"\nFunction invocation failed due to no matching argument types."
)
diagnosis_ambiguous_encoding = (
    r"\nAmbiguous argument encoding. "
    r"Provided arguments can be encoded to multiple functions matching this call."
)


def test_no_functions_match_identifier(arrays_contract):
    with pytest.raises(MismatchedABI):
        arrays_contract.functions.thisFunctionDoesNotExist().call()


def test_function_1_match_identifier_wrong_number_of_args(arrays_contract):
    regex = message_regex + diagnosis_arg_regex
    with pytest.raises(ValidationError, match=regex):
        arrays_contract.functions.setBytes32Value().call()


def test_function_1_match_identifier_wrong_args_encoding(arrays_contract):
    regex = message_regex + diagnosis_encoding_regex
    with pytest.raises(ValidationError, match=regex):
        arrays_contract.functions.setBytes32Value('dog').call()


def test_function_multiple_match_identifiers_no_correct_number_of_args(web3):
    MULTIPLE_FUNCTIONS = json.loads('[{"constant":false,"inputs":[],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"bytes32"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint256"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint8"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"int8"}],"name":"a","outputs":[],"type":"function"}]')  # noqa: E501
    Contract = web3.eth.contract(abi=MULTIPLE_FUNCTIONS)
    regex = message_regex + diagnosis_arg_regex
    with pytest.raises(ValidationError, match=regex):
        Contract.functions.a(100, 'dog').call()


def test_function_multiple_match_identifiers_no_correct_encoding_of_args(web3):
    MULTIPLE_FUNCTIONS = json.loads('[{"constant":false,"inputs":[],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"bytes32"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint256"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint8"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"int8"}],"name":"a","outputs":[],"type":"function"}]')  # noqa: E501
    Contract = web3.eth.contract(abi=MULTIPLE_FUNCTIONS)
    regex = message_regex + diagnosis_encoding_regex
    with pytest.raises(ValidationError, match=regex):
        Contract.functions.a('dog').call()


def test_function_multiple_possible_encodings(web3):
    MULTIPLE_FUNCTIONS = json.loads('[{"constant":false,"inputs":[],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"bytes32"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint256"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"uint8"}],"name":"a","outputs":[],"type":"function"},{"constant":false,"inputs":[{"name":"","type":"int8"}],"name":"a","outputs":[],"type":"function"}]')  # noqa: E501
    Contract = web3.eth.contract(abi=MULTIPLE_FUNCTIONS)
    regex = message_regex + diagnosis_ambiguous_encoding
    with pytest.raises(ValidationError, match=regex):
        Contract.functions.a(100).call()


def test_function_no_abi(web3):
    contract = web3.eth.contract()
    with pytest.raises(NoABIFunctionsFound):
        contract.functions.thisFunctionDoesNotExist().call()


def test_call_abi_no_functions(web3):
    contract = web3.eth.contract(abi=[])
    with pytest.raises(NoABIFunctionsFound):
        contract.functions.thisFunctionDoesNotExist().call()


@pytest.mark.parametrize(
    'function, value',
    (
        # minimum positive unambiguous value (larger than fixed8x1)
        ('reflect', Decimal('12.8')),
        # maximum value (for ufixed256x1)
        ('reflect', Decimal(2 ** 256 - 1) / 10),
        # maximum negative unambiguous value (less than 0 from ufixed*)
        ('reflect', Decimal('-0.1')),
        # minimum value (for fixed8x1)
        ('reflect', Decimal('-12.8')),
        # only ufixed256x80 type supports 2-80 decimals
        ('reflect', Decimal(2 ** 256 - 1) / 10 ** 80),  # maximum allowed value
        ('reflect', Decimal(1) / 10 ** 80),  # smallest non-zero value
        # minimum value (for ufixed8x1)
        ('reflect_short_u', 0),
        # maximum value (for ufixed8x1)
        ('reflect_short_u', Decimal('25.5')),
    ),
)
def test_reflect_fixed_value(web3, fixed_reflection_contract, function, value):
    contract_func = fixed_reflection_contract.functions[function]
    reflected = contract_func(value).call({'gas': 420000})
    assert reflected == value


DEFAULT_DECIMALS = getcontext().prec


@pytest.mark.parametrize(
    'function, value, error',
    (
        # out of range
        ('reflect_short_u', Decimal('25.6'), "no matching argument types"),
        ('reflect_short_u', Decimal('-.1'), "no matching argument types"),
        # too many digits for *x1, too large for 256x80
        ('reflect', Decimal('0.01'), "no matching argument types"),

        # too many digits
        ('reflect_short_u', Decimal('0.01'), "no matching argument types"),
        (
            'reflect_short_u',
            Decimal('1e-%d' % (DEFAULT_DECIMALS + 1)),
            "no matching argument types",
        ),
        ('reflect_short_u', Decimal('25.4' + '9' * DEFAULT_DECIMALS), "no matching argument types"),
        ('reflect', Decimal(1) / 10 ** 81, "no matching argument types"),

        # floats not accepted, for floating point error concerns
        ('reflect_short_u', 0.1, "no matching argument types"),

        # ambiguous
        ('reflect', Decimal('12.7'), "Ambiguous argument encoding"),
        ('reflect', Decimal(0), "Ambiguous argument encoding"),
        ('reflect', 0, "Ambiguous argument encoding"),
    ),
)
def test_invalid_fixed_value_reflections(web3, fixed_reflection_contract, function, value, error):
    contract_func = fixed_reflection_contract.functions[function]
    with pytest.raises(ValidationError, match=error):
        contract_func(value).call({'gas': 420000})



import pytest

from eth_utils import (
    is_address,
)

# Ignore warning in pyethereum 1.6 - will go away with the upgrade
pytestmark = pytest.mark.filterwarnings("ignore:implicit cast from 'char *'")


@pytest.mark.parametrize('call_as_instance', (True, False))
def test_create_filter_address_parameter(web3, emitter, Emitter, call_as_instance):
    if call_as_instance:
        event_filter = emitter.events.LogNoArguments.createFilter(fromBlock="latest")
    else:
        event_filter = Emitter.events.LogNoArguments.createFilter(fromBlock="latest")

    if call_as_instance:
        # Assert this is a single string value, and not a list of addresses
        assert is_address(event_filter.filter_params['address'])
    else:
        #  Undeployed contract shouldnt have address...
        assert 'address' not in event_filter.filter_params


@pytest.mark.parametrize('call_as_instance', (True, False))
def test_on_filter_using_get_entries_interface(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter):

    if call_as_instance:
        event_filter = create_filter(emitter, ['LogNoArguments', {}])
    else:
        event_filter = create_filter(Emitter, ['LogNoArguments', {}])

    txn_hash = emitter.functions.logNoArgs(emitter_event_ids.LogNoArguments).transact()
    wait_for_transaction(web3, txn_hash)

    log_entries = event_filter.get_new_entries()
    assert len(log_entries) == 1
    assert log_entries[0]['transactionHash'] == txn_hash

    # a second call is empty because all events have been retrieved
    new_entries = event_filter.get_new_entries()
    assert len(new_entries) == 0


@pytest.mark.parametrize('call_as_instance', (True, False))
def test_on_sync_filter_with_event_name_and_single_argument(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter):

    if call_as_instance:
        event_filter = create_filter(emitter, ['LogTripleWithIndex', {'filter': {
            'arg1': 2,
        }}])
    else:
        event_filter = create_filter(Emitter, ['LogTripleWithIndex', {'filter': {
            'arg1': 2,
        }}])

    txn_hashes = []
    event_id = emitter_event_ids.LogTripleWithIndex
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 2, 1, 3).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 2, 3).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 12345, 2, 54321).transact()
    )
    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    seen_logs = event_filter.get_new_entries()
    assert len(seen_logs) == 2
    assert {l['transactionHash'] for l in seen_logs} == set(txn_hashes[1:])


@pytest.mark.parametrize('call_as_instance', (True, False))
def test_on_sync_filter_with_event_name_and_non_indexed_argument(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter):

    if call_as_instance:
        contract = emitter
    else:
        contract = Emitter

    event_filter = create_filter(contract, ['LogTripleWithIndex', {'filter': {
        'arg0': 1,
        'arg1': 2,
    }}])

    txn_hashes = []
    event_id = emitter_event_ids.LogTripleWithIndex
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 2, 1, 3).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 2, 3).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 12345, 2, 54321).transact()
    )
    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    seen_logs = event_filter.get_new_entries()
    assert len(seen_logs) == 1
    assert seen_logs[0]['transactionHash'] == txn_hashes[1]

    post_event_filter = contract.events.LogTripleWithIndex.createFilter(
        argument_filters={'arg0': 1, 'arg1': 2},
        fromBlock=0,
    )

    old_logs = post_event_filter.get_all_entries()
    assert len(old_logs) == 1
    assert old_logs[0]['transactionHash'] == txn_hashes[1]


def test_filter_with_contract_address(web3, emitter, emitter_event_ids, wait_for_transaction):
    event_filter = web3.eth.filter(filter_params={'address': emitter.address})
    txn_hash = emitter.functions.logNoArgs(emitter_event_ids.LogNoArguments).transact()
    wait_for_transaction(web3, txn_hash)
    seen_logs = event_filter.get_new_entries()
    assert len(seen_logs) == 1
    assert seen_logs[0]['transactionHash'] == txn_hash


@pytest.mark.parametrize('call_as_instance', (True, False))
def test_on_sync_filter_with_topic_filter_options(
        web3,
        emitter,
        Emitter,
        wait_for_transaction,
        emitter_event_ids,
        call_as_instance,
        create_filter):

    if call_as_instance:
        contract = emitter
    else:
        contract = Emitter

    event_filter = create_filter(contract, ['LogTripleWithIndex', {'filter': {
        'arg1': [1, 2],
        'arg2': [1, 2]
    }}])

    txn_hashes = []
    event_id = emitter_event_ids.LogTripleWithIndex
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 1, 1).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 1, 2).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 2, 2).transact()
    )
    txn_hashes.append(
        emitter.functions.logTriple(event_id, 1, 2, 1).transact()
    )
    for txn_hash in txn_hashes:
        wait_for_transaction(web3, txn_hash)

    seen_logs = event_filter.get_new_entries()
    assert len(seen_logs) == 4

    post_event_filter = contract.events.LogTripleWithIndex.createFilter(
        argument_filters={'arg1': [1, 2], 'arg2': [1, 2]},
        fromBlock=0,
    )

    old_logs = post_event_filter.get_all_entries()
    assert len(old_logs) == 4



import pytest


@pytest.fixture(autouse=True)
def wait_for_first_block(web3, wait_for_block):
    wait_for_block(web3)


def test_uses_defaultAccount_when_set(web3, extra_accounts,
                                      wait_for_transaction):
    web3.eth.defaultAccount = extra_accounts[2]

    txn_hash = web3.eth.sendTransaction({
        "to": extra_accounts[1],
        "value": 1234,
    })

    wait_for_transaction(web3, txn_hash)

    txn = web3.eth.getTransaction(txn_hash)
    assert txn['from'] == extra_accounts[2]


def test_uses_given_from_address_when_provided(web3, extra_accounts,
                                               wait_for_transaction):
    web3.eth.defaultAccount = extra_accounts[2]
    txn_hash = web3.eth.sendTransaction({
        "from": extra_accounts[5],
        "to": extra_accounts[1],
        "value": 1234,
    })

    wait_for_transaction(web3, txn_hash)

    txn = web3.eth.getTransaction(txn_hash)
    assert txn['from'] == extra_accounts[5]



import pytest

from web3.exceptions import (
    MismatchedABI,
    NoABIEventsFound,
)

EVENT_1_ABI = {
    "anonymous": False,
    "inputs": [
        {"indexed": False, "name": "arg0", "type": "uint256"},
        {"indexed": True, "name": "arg1", "type": "uint256"},
    ],
    "name": "Event_1",
    "type": "event",
}


def test_access_event_with_no_abi(web3):
    contract = web3.eth.contract()
    with pytest.raises(NoABIEventsFound):
        contract.events.thisEventDoesNotExist()


def test_access_event_abi_with_no_events(web3):
    contract = web3.eth.contract(abi=[])
    with pytest.raises(NoABIEventsFound):
        contract.events.thisEventDoesNotExist()


def test_access_nonexistent_event(web3):
    contract = web3.eth.contract(abi=[EVENT_1_ABI])
    with pytest.raises(MismatchedABI):
        contract.events.thisEventDoesNotExist()



import pytest

from web3.datastructures import (
    AttributeDict,
)


@pytest.mark.parametrize(
    "dict1,dict2,same_hash",
    [
        ({'b': 2, 'a': 1}, {'a': 1, 'b': 2}, True),
        ({'a': 1, 'b': 2}, {'a': 1, 'b': 0}, False),
    ],
)
def test_attributedict_hashable(dict1, dict2, same_hash):
    dicts = map(AttributeDict, (dict1, dict2))
    dicts = set(dicts)
    assert len(dicts) == (1 if same_hash else 2)


def test_attributedict_access():
    container = AttributeDict({'a': 1})
    assert container.a == 1


def test_attributedict_repr():
    dict1 = AttributeDict({'a': 1})
    dict2 = eval(repr(dict1))
    assert dict1 == dict2


def test_attributedict_setitem_invalid():
    container = AttributeDict({'a': 1})
    with pytest.raises(TypeError):
        container['a'] = 0
    assert container['a'] == 1


def test_attributedict_setattr_invalid():
    container = AttributeDict({'a': 1})
    with pytest.raises(TypeError):
        container.a = 0
    assert container.a == 1


def test_attributedict_delitem_invalid():
    container = AttributeDict({'a': 1})
    with pytest.raises(TypeError):
        del container['a']
    assert container['a'] == 1


@pytest.mark.parametrize(
    "dict1,dict2",
    [
        ({'b': 2, 'a': 1}, {'a': 1, 'b': 2}),
        ({}, {}),
    ],
)
def test_attributedict_equality(dict1, dict2):
    assert AttributeDict(dict1) == dict2
    assert AttributeDict(dict1) == AttributeDict(dict2)
    assert dict1 == AttributeDict(dict2)


@pytest.mark.parametrize(
    "dict1,dict2",
    [
        ({'a': 1, 'b': 2}, {'a': 1, 'b': 0}),
        ({'a': 1, 'b': 2}, {'a': 1}),
    ],
)
def test_attributedict_inequality(dict1, dict2):
    assert AttributeDict(dict1) != dict2
    assert AttributeDict(dict1) != AttributeDict(dict2)
    assert dict1 != AttributeDict(dict2)


def test_attributedict_recursive_dict():
    w = AttributeDict.recursive({'x': {'y': {'z': 8}}})
    assert w.x.y.z == 8


@pytest.mark.parametrize('sequence', [list, tuple])
def test_attributedict_sequence_with_dict(sequence):
    data = sequence(['a', {'found': True}, 'c'])
    dict_in_sequence = AttributeDict.recursive(data)
    assert dict_in_sequence[1].found is True


def test_attributedict_dict_in_list_in_dict():
    data = {'instructions': [
        0,
        1,
        'neither shalt thou count, excepting that thou then proceedeth to three',
        {'if_naughty': 'snuff it'},
        'shalt thou not count',
        'right out',
    ]}
    attrdict = AttributeDict.recursive(data)
    assert attrdict.instructions[3].if_naughty == 'snuff it'


def test_attributedict_set_in_recursive_dict():
    data = {'mydict': {'myset': {'found'}}}
    attrdict = AttributeDict.recursive(data)
    assert 'found' in attrdict.mydict.myset




import pytest

from eth_utils import (
    decode_hex,
)

# Ignore warning in pyethereum 1.6 - will go away with the upgrade
pytestmark = pytest.mark.filterwarnings("ignore:implicit cast from 'char *'")


def test_contract_deployment_no_constructor(web3, MathContract,
                                            MATH_RUNTIME):
    deploy_txn = MathContract.constructor().transact()

    txn_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert txn_receipt is not None

    assert txn_receipt['contractAddress']
    contract_address = txn_receipt['contractAddress']

    blockchain_code = web3.eth.getCode(contract_address)
    assert blockchain_code == decode_hex(MATH_RUNTIME)


def test_contract_deployment_with_constructor_without_args(web3,
                                                           SimpleConstructorContract,
                                                           SIMPLE_CONSTRUCTOR_RUNTIME):
    deploy_txn = SimpleConstructorContract.constructor().transact()

    txn_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert txn_receipt is not None

    assert txn_receipt['contractAddress']
    contract_address = txn_receipt['contractAddress']

    blockchain_code = web3.eth.getCode(contract_address)
    assert blockchain_code == decode_hex(SIMPLE_CONSTRUCTOR_RUNTIME)


def test_contract_deployment_with_constructor_with_arguments(web3,
                                                             WithConstructorArgumentsContract,
                                                             WITH_CONSTRUCTOR_ARGUMENTS_RUNTIME):
    deploy_txn = WithConstructorArgumentsContract.constructor(1234, 'abcd').transact()

    txn_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert txn_receipt is not None

    assert txn_receipt['contractAddress']
    contract_address = txn_receipt['contractAddress']

    blockchain_code = web3.eth.getCode(contract_address)
    assert blockchain_code == decode_hex(WITH_CONSTRUCTOR_ARGUMENTS_RUNTIME)


def test_contract_deployment_with_constructor_with_address_argument(web3,
                                                                    WithConstructorAddressArgumentsContract,  # noqa: E501
                                                                    WITH_CONSTRUCTOR_ADDRESS_RUNTIME):  # noqa: E501
    deploy_txn = WithConstructorAddressArgumentsContract.constructor(
        "0x16D9983245De15E7A9A73bC586E01FF6E08dE737",
    ).transact()

    txn_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert txn_receipt is not None

    assert txn_receipt['contractAddress']
    contract_address = txn_receipt['contractAddress']

    blockchain_code = web3.eth.getCode(contract_address)
    assert blockchain_code == decode_hex(WITH_CONSTRUCTOR_ADDRESS_RUNTIME)



import pytest

from web3.utils.transactions import (
    assert_valid_transaction_params,
    extract_valid_transaction_params,
)


def test_assert_valid_transaction_params_all_params():
    assert_valid_transaction_params({
        'from': '0x0',
        'to': '0x0',
        'gas': 21000,
        'gasPrice': 5000000,
        'value': 1,
        'data': '0x0',
        'nonce': 2,
        'chainId': 1,
    })


def test_assert_valid_transaction_params_some_params():
    assert_valid_transaction_params({
        'from': '0x0',
        'to': '0x0',
        'value': 1,
    })


def test_assert_valid_transaction_params_invalid_param():
    with pytest.raises(ValueError):
        assert_valid_transaction_params({
            'from': '0x0',
            'to': '0x0',
            'value': 1,
            'tokens': 9000,
        })


FULL_TXN_DICT = {
    'from': '0x0',
    'to': '0x1',
    'gas': 21000,
    'gasPrice': 5000000,
    'data': '0x2',
    'value': 3,
    'nonce': 2,
    'chainId': 1,
}


@pytest.mark.parametrize(
    "transaction_params, expected",
    ((FULL_TXN_DICT, FULL_TXN_DICT),
     ({'data': '0x0', 'input': '0x0'}, {'data': '0x0'}),
     ({'input': '0x0'}, {'data': '0x0'}),
     ({}, {}),
     )
)
def test_extract_valid_transaction_params(transaction_params, expected):
    valid_transaction_params = extract_valid_transaction_params(transaction_params)
    assert valid_transaction_params == expected


INVALID_TXN_PARAMS = {'data': '0x0', 'input': '0x1'}
EXPECTED_EXC_MSG = r'.* "input:(.*)" and "data:(.*)" .*'


@pytest.mark.parametrize(
    "transaction_params, expected_exc_msg",
    ((INVALID_TXN_PARAMS, EXPECTED_EXC_MSG),)
)
def test_extract_valid_transaction_params_invalid(transaction_params, expected_exc_msg):
    with pytest.raises(AttributeError, match=expected_exc_msg):
        extract_valid_transaction_params(transaction_params)


def test_extract_valid_transaction_params_includes_invalid():
    input = {
        'from': '0x0',
        'to': '0x0',
        'gas': 21000,
        'gasPrice': 5000000,
        'value': 1,
        'tokens': 9000,
    }
    valid_transaction_params = extract_valid_transaction_params(input)
    assert valid_transaction_params == {
        'from': '0x0',
        'to': '0x0',
        'gas': 21000,
        'gasPrice': 5000000,
        'value': 1,
    }



import random

from flaky import (
    flaky,
)

from web3.utils.threads import (
    Timeout,
)


@flaky(max_runs=3)
def test_miner_setGasPrice(web3_empty, wait_for_block):
    web3 = web3_empty

    initial_gas_price = web3.eth.gasPrice

    # sanity check
    assert web3.eth.gasPrice > 1000

    web3.miner.setGasPrice(initial_gas_price // 2)

    with Timeout(60) as timeout:
        while web3.eth.gasPrice == initial_gas_price:
            timeout.sleep(random.random())

    after_gas_price = web3.eth.gasPrice
    assert after_gas_price < initial_gas_price



import pytest

from web3.exceptions import (
    BadFunctionCallOutput,
    NameNotFound,
)
from web3.utils.ens import (
    contract_ens_addresses,
    ens_addresses,
)


@pytest.fixture()
def math_addr(MathContract, address_conversion_func):
    web3 = MathContract.web3
    deploy_txn = MathContract.constructor().transact({'from': web3.eth.coinbase})
    deploy_receipt = web3.eth.waitForTransactionReceipt(deploy_txn)
    assert deploy_receipt is not None
    return address_conversion_func(deploy_receipt['contractAddress'])


def test_contract_with_unset_address(MathContract):
    with contract_ens_addresses(MathContract, []):
        with pytest.raises(NameNotFound):
            MathContract(address='unsetname.eth')


def test_contract_with_name_address(MathContract, math_addr):
    with contract_ens_addresses(MathContract, [('thedao.eth', math_addr)]):
        mc = MathContract(address='thedao.eth')
        caller = mc.web3.eth.coinbase
        assert mc.address == 'thedao.eth'
        assert mc.functions.return13().call({'from': caller}) == 13


def test_contract_with_name_address_from_eth_contract(
    web3,
    MATH_ABI,
    MATH_CODE,
    MATH_RUNTIME,
    math_addr,
):
    with ens_addresses(web3, [('thedao.eth', math_addr)]):
        mc = web3.eth.contract(
            address='thedao.eth',
            abi=MATH_ABI,
            bytecode=MATH_CODE,
            bytecode_runtime=MATH_RUNTIME,
        )

        caller = mc.web3.eth.coinbase
        assert mc.address == 'thedao.eth'
        assert mc.functions.return13().call({'from': caller}) == 13


def test_contract_with_name_address_changing(MathContract, math_addr):
    # Contract address is validated once on creation
    with contract_ens_addresses(MathContract, [('thedao.eth', math_addr)]):
        mc = MathContract(address='thedao.eth')

    caller = mc.web3.eth.coinbase
    assert mc.address == 'thedao.eth'

    # what happen when name returns no address at all
    with contract_ens_addresses(mc, []):
        with pytest.raises(NameNotFound):
            mc.functions.return13().call({'from': caller})

    # what happen when name returns address to different contract
    with contract_ens_addresses(mc, [('thedao.eth', '0x' + '11' * 20)]):
        with pytest.raises(BadFunctionCallOutput):
            mc.functions.return13().call({'from': caller})

    # contract works again when name resolves correctly
    with contract_ens_addresses(mc, [('thedao.eth', math_addr)]):
        assert mc.functions.return13().call({'from': caller}) == 13



import os
import pytest
import shutil
import tempfile

from eth_utils import (
    is_checksum_address,
    is_dict,
)

from .install_parity import (
    get_executable_path,
    install_parity,
)
from .utils import (
    get_process,
)

KEYFILE_PW = 'web3py-test'

PARITY_1_8_8_FIXTURE = {
    'datadir': 'parity-188-fixture',
    'coinbase': 'dc544d1aa88ff8bbd2f2aec754b1f1e99e1812fd',
    'block_hash_with_log': '0xf9aea1029be4b93e30313814d297a2da0c5b204856c4de16c21b87ebaf8f20c5',
    'block_with_txn_hash': '0xf460bffb4ea46c5ec26e2062911f24af3349cd1c895fdb009b4d544569d22a49',
    'emitter_address': '0x4aA591a07989b4F810E2F5cE97e769D60710f168',
    'emitter_deploy_txn_hash': '0xa81e903e9953758c8da5aaae66451ff909edd7bd6aefc3ebeab1e709e3229bcc',
    'empty_block_hash': '0x3b32b087d4fc72091be4fa6b44fc2a78289a15c53fba3454755dd9eaf39f0481',
    'keyfile_pw': 'web3py-test',
    'math_address': '0xd794C821fCCFF5D96F5Db44af7e29977630A9dc2',
    'math_deploy_txn_hash': '0x03cc47c8f58608576187825aed01c4fc64786f1172d182d432336881a75a0fa3',
    'mined_txn_hash': '0x9839fde5fce7f0ed29b49a687d4f7630076069e65c2e1df87ffab9b2844d3899',
    'raw_txn_account': '0x39EEed73fb1D3855E90Cbd42f348b3D7b340aAA6',
    'txn_hash_with_log': '0x26bad3318b3466833f96d04ac9ba46fbbce11c15be2f83c9fe0b5dc15b2646cd'
}

PARITY_1_8_7_FIXTURE = {
    'datadir': 'parity-187-fixture',
    'coinbase': 'dc544d1aa88ff8bbd2f2aec754b1f1e99e1812fd',
    'block_hash_with_log': '0x342e12ab6d24d7fb1d774a6b47cd2cc04430a3295bee5662d5a1a0b766480031',
    'block_with_txn_hash': '0xa866266a5a348948c38855cc6e990093b35a3d2c43fdddfe3b1259c9c3fc7404',
    'emitter_address': '0x4aA591a07989b4F810E2F5cE97e769D60710f168',
    'emitter_deploy_txn_hash': '0xa81e903e9953758c8da5aaae66451ff909edd7bd6aefc3ebeab1e709e3229bcc',
    'empty_block_hash': '0xbcb2826e4376c23e66750607af72965f177f93b39e5024be259e6b0ff4f95e9d',
    'keyfile_pw': 'web3py-test',
    'math_address': '0xd794C821fCCFF5D96F5Db44af7e29977630A9dc2',
    'math_deploy_txn_hash': '0x03cc47c8f58608576187825aed01c4fc64786f1172d182d432336881a75a0fa3',
    'mined_txn_hash': '0x9839fde5fce7f0ed29b49a687d4f7630076069e65c2e1df87ffab9b2844d3899',
    'raw_txn_account': '0x39EEed73fb1D3855E90Cbd42f348b3D7b340aAA6',
    'txn_hash_with_log': '0x26bad3318b3466833f96d04ac9ba46fbbce11c15be2f83c9fe0b5dc15b2646cd'
}


@pytest.fixture(scope='module')
def parity_binary():
    if 'PARITY_BINARY' in os.environ:
        return os.environ['PARITY_BINARY']
    elif 'PARITY_VERSION' in os.environ:
        parity_version = os.environ['PARITY_VERSION']
        _parity_binary = get_executable_path(parity_version)
        if not os.path.exists(_parity_binary):
            install_parity(parity_version)
        assert os.path.exists(_parity_binary)
        return _parity_binary
    else:
        return 'parity'


def get_parity_version(parity_binary):
    pass


@pytest.fixture(scope="module")
def parity_fixture_data(parity_binary):
    if os.path.basename(parity_binary) == 'parity-1_8_7':
        return PARITY_1_8_7_FIXTURE
    elif os.path.basename(parity_binary) == 'parity-1_8_8':
        return PARITY_1_8_8_FIXTURE
    else:
        return PARITY_1_8_7_FIXTURE


@pytest.fixture(scope='module')
def datadir(tmpdir_factory, parity_fixture_data):
    fixture_datadir = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        '..',
        parity_fixture_data['datadir'],
    ))
    base_dir = tmpdir_factory.mktemp('parity')
    tmp_datadir = os.path.join(str(base_dir), 'datadir')
    shutil.copytree(fixture_datadir, tmp_datadir)
    return tmp_datadir


@pytest.fixture(scope="module")
def author(parity_fixture_data):
    # need the address to unlock before web3 module has been opened
    author = parity_fixture_data['coinbase']
    return author


@pytest.fixture(scope="module")
def passwordfile():
    password_dir = tempfile.mkdtemp()
    password_path = os.path.join(password_dir, 'password')
    with open(password_path, 'w') as f:
        f.write(KEYFILE_PW)

    yield password_path

    if os.path.exists(password_path):
        os.remove(password_path)


@pytest.fixture(scope="module")
def parity_process(parity_command_arguments):
    yield from get_process(parity_command_arguments)


@pytest.fixture(scope="module")
def parity_import_blocks_process(parity_import_blocks_command):
    yield from get_process(parity_import_blocks_command, terminates=True)


@pytest.fixture(scope='module')
def coinbase(web3):
    return web3.eth.coinbase


@pytest.fixture(scope="module")
def math_contract_deploy_txn_hash(parity_fixture_data):
    return parity_fixture_data['math_deploy_txn_hash']


@pytest.fixture(scope="module")
def math_contract(web3, math_contract_factory, parity_fixture_data):
    return math_contract_factory(address=parity_fixture_data['math_address'])


@pytest.fixture()
def math_contract_address(math_contract, address_conversion_func):
    return address_conversion_func(math_contract.address)


@pytest.fixture(scope="module")
def emitter_contract(web3, emitter_contract_factory, parity_fixture_data):
    return emitter_contract_factory(address=parity_fixture_data['emitter_address'])


@pytest.fixture()
def emitter_contract_address(emitter_contract, address_conversion_func):
    return address_conversion_func(emitter_contract.address)


@pytest.fixture(scope="module")
def unlocked_account(web3, unlockable_account, unlockable_account_pw):
    yield unlockable_account


@pytest.fixture(scope='module')
def unlockable_account_pw(parity_fixture_data):
    return parity_fixture_data['keyfile_pw']


@pytest.fixture(scope="module")
def unlockable_account(web3, coinbase):
    yield coinbase


@pytest.fixture()
def unlockable_account_dual_type(unlockable_account, address_conversion_func):
    return address_conversion_func(unlockable_account)


@pytest.fixture
def unlocked_account_dual_type(unlockable_account_dual_type):
    return unlockable_account_dual_type


@pytest.fixture(scope="module")
def funded_account_for_raw_txn(parity_fixture_data):
    account = parity_fixture_data['raw_txn_account']
    assert is_checksum_address(account)
    return account


@pytest.fixture(scope="module")
def empty_block(web3, parity_fixture_data):
    block = web3.eth.getBlock(parity_fixture_data['empty_block_hash'])
    assert is_dict(block)
    return block


@pytest.fixture(scope="module")
def block_with_txn(web3, parity_fixture_data):
    block = web3.eth.getBlock(parity_fixture_data['block_with_txn_hash'])
    assert is_dict(block)
    return block


@pytest.fixture(scope="module")
def mined_txn_hash(parity_fixture_data):
    return parity_fixture_data['mined_txn_hash']


@pytest.fixture(scope="module")
def block_with_txn_with_log(web3, parity_fixture_data):
    block = web3.eth.getBlock(parity_fixture_data['block_hash_with_log'])
    assert is_dict(block)
    return block


@pytest.fixture(scope="module")
def txn_hash_with_log(parity_fixture_data):
    return parity_fixture_data['txn_hash_with_log']


@pytest.fixture(scope="module")
def txn_filter_params(coinbase):
    return {
        "fromBlock": "earliest",
        "toBlock": "latest",
        "fromAddress": [coinbase],
    }



import pytest

from tests.integration.common import (
    MiscWebsocketTest,
)
from tests.integration.utils import (
    wait_for_ws,
)
from web3 import Web3

from .common import (
    GoEthereumEthModuleTest,
    GoEthereumNetModuleTest,
    GoEthereumPersonalModuleTest,
    GoEthereumTest,
    GoEthereumVersionModuleTest,
    get_open_port,
)


@pytest.fixture(scope="module")
def ws_port():
    return get_open_port()


@pytest.fixture(scope="module")
def endpoint_uri(ws_port):
    return 'ws://localhost:{0}'.format(ws_port)


@pytest.fixture(scope='module')
def geth_command_arguments(geth_binary, datadir, ws_port):
    return (
        geth_binary,
        '--datadir', str(datadir),
        '--nodiscover',
        '--fakepow',
        '--ws',
        '--wsport', ws_port,
        '--wsapi', 'db,eth,net,web3,personal,web3',
        '--wsorigins', '*',
        '--ipcdisable',
    )


@pytest.fixture(scope="module")
def web3(geth_process, endpoint_uri, event_loop):
    event_loop.run_until_complete(wait_for_ws(endpoint_uri, event_loop))
    _web3 = Web3(Web3.WebsocketProvider(endpoint_uri))
    return _web3


class TestGoEthereumTest(GoEthereumTest):
    pass


class TestGoEthereumEthModuleTest(GoEthereumEthModuleTest):
    pass


class TestGoEthereumVersionModuleTest(GoEthereumVersionModuleTest):
    pass


class TestGoEthereumNetModuleTest(GoEthereumNetModuleTest):
    pass


class TestGoEthereumPersonalModuleTest(GoEthereumPersonalModuleTest):
    pass


class TestMiscWebsocketTest(MiscWebsocketTest):
    pass



from eth_utils import (
    apply_to_return_value,
    add_0x_prefix,
    from_wei,
    is_address,
    is_checksum_address,
    keccak,
    remove_0x_prefix,
    to_checksum_address,
    to_wei,
)

from ens import ENS

from web3.admin import Admin
from web3.eth import Eth
from web3.iban import Iban
from web3.miner import Miner
from web3.net import Net
from web3.parity import Parity
from web3.personal import Personal
from web3.testing import Testing
from web3.txpool import TxPool
from web3.version import Version

from web3.providers.eth_tester import (
    EthereumTesterProvider,
)
from web3.providers.ipc import (
    IPCProvider,
)
from web3.providers.rpc import (
    HTTPProvider,
)
from web3.providers.tester import (
    TestRPCProvider,
)
from web3.providers.websocket import (
    WebsocketProvider
)

from web3.manager import (
    RequestManager,
)

from web3.utils.abi import (
    map_abi_data,
)
from hexbytes import (
    HexBytes,
)
from web3.utils.decorators import (
    combomethod,
)
from web3.utils.empty import empty
from web3.utils.encoding import (
    hex_encode_abi_type,
    to_bytes,
    to_int,
    to_hex,
    to_text,
)
from web3.utils.normalizers import (
    abi_ens_resolver,
)


def get_default_modules():
    return {
        "eth": Eth,
        "net": Net,
        "personal": Personal,
        "version": Version,
        "txpool": TxPool,
        "miner": Miner,
        "admin": Admin,
        "parity": Parity,
        "testing": Testing,
    }


class Web3:
    # Providers
    HTTPProvider = HTTPProvider
    IPCProvider = IPCProvider
    TestRPCProvider = TestRPCProvider
    EthereumTesterProvider = EthereumTesterProvider
    WebsocketProvider = WebsocketProvider

    # Managers
    RequestManager = RequestManager

    # Iban
    Iban = Iban

    # Encoding and Decoding
    toBytes = staticmethod(to_bytes)
    toInt = staticmethod(to_int)
    toHex = staticmethod(to_hex)
    toText = staticmethod(to_text)

    # Currency Utility
    toWei = staticmethod(to_wei)
    fromWei = staticmethod(from_wei)

    # Address Utility
    isAddress = staticmethod(is_address)
    isChecksumAddress = staticmethod(is_checksum_address)
    toChecksumAddress = staticmethod(to_checksum_address)

    def __init__(self, providers=empty, middlewares=None, modules=None, ens=empty):
        self.manager = RequestManager(self, providers, middlewares)

        if modules is None:
            modules = get_default_modules()

        for module_name, module_class in modules.items():
            module_class.attach(self, module_name)

        self.ens = ens

    @property
    def middleware_stack(self):
        return self.manager.middleware_stack

    @property
    def providers(self):
        return self.manager.providers

    @providers.setter
    def providers(self, providers):
        self.manager.providers = providers

    @staticmethod
    @apply_to_return_value(HexBytes)
    def sha3(primitive=None, text=None, hexstr=None):
        if isinstance(primitive, (bytes, int, type(None))):
            input_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
            return keccak(input_bytes)

        raise TypeError(
            "You called sha3 with first arg %r and keywords %r. You must call it with one of "
            "these approaches: sha3(text='txt'), sha3(hexstr='0x747874'), "
            "sha3(b'\\x74\\x78\\x74'), or sha3(0x747874)." % (
                primitive,
                {'text': text, 'hexstr': hexstr}
            )
        )

    @combomethod
    def soliditySha3(cls, abi_types, values):
        """
        Executes sha3 (keccak256) exactly as Solidity does.
        Takes list of abi_types as inputs -- `[uint24, int8[], bool]`
        and list of corresponding values  -- `[20, [-1, 5, 0], True]`
        """
        if len(abi_types) != len(values):
            raise ValueError(
                "Length mismatch between provided abi types and values.  Got "
                "{0} types and {1} values.".format(len(abi_types), len(values))
            )

        if isinstance(cls, type):
            w3 = None
        else:
            w3 = cls
        normalized_values = map_abi_data([abi_ens_resolver(w3)], abi_types, values)

        hex_string = add_0x_prefix(''.join(
            remove_0x_prefix(hex_encode_abi_type(abi_type, value))
            for abi_type, value
            in zip(abi_types, normalized_values)
        ))
        return cls.sha3(hexstr=hex_string)

    def isConnected(self):
        for provider in self.providers:
            if provider.isConnected():
                return True
        else:
            return False

    @property
    def ens(self):
        if self._ens is empty:
            return ENS.fromWeb3(self)
        else:
            return self._ens

    @ens.setter
    def ens(self, new_ens):
        self._ens = new_ens



from unittest.mock import (
    Mock,
)


def test_no_gas_price_strategy_returns_none(web3):
    assert web3.eth.generateGasPrice() is None


def test_set_gas_price_strategy(web3):
    def my_gas_price_strategy(web3, transaction_params):
        return 5
    web3.eth.setGasPriceStrategy(my_gas_price_strategy)
    assert web3.eth.generateGasPrice() == 5


def test_gas_price_strategy_calls(web3):
    transaction = {
        'to': '0x0',
        'value': 1000000000
    }
    my_gas_price_strategy = Mock(return_value=5)
    web3.eth.setGasPriceStrategy(my_gas_price_strategy)
    assert web3.eth.generateGasPrice(transaction) == 5
    my_gas_price_strategy.assert_called_once_with(web3, transaction)



import pytest

from web3 import Web3
from web3.middleware import (  # noqa: F401
    construct_result_generator_middleware,
    request_parameter_normalizer,
)
from web3.providers.base import (
    BaseProvider,
)


@pytest.fixture
def w3_base():
    return Web3(providers=[BaseProvider()], middlewares=[])


@pytest.fixture
def result_generator_middleware():
    return construct_result_generator_middleware({
        'eth_getLogs': lambda _, params: params,
    })


@pytest.fixture
def w3(w3_base, result_generator_middleware):
    w3_base.middleware_stack.add(result_generator_middleware)
    w3_base.middleware_stack.add(request_parameter_normalizer)
    return w3_base


def test_eth_getLogs_param_normalization(w3):
    result = w3.eth.getLogs({
        'from': 'latest', 'address': '0x1111111111111111111111111111111111111111'})
    assert isinstance(result[0]['address'], list)



import pytest

from web3 import Web3
from web3.exceptions import (
    ValidationError,
)
from web3.gas_strategies.time_based import (
    construct_time_based_gas_price_strategy,
)
from web3.middleware import (
    construct_result_generator_middleware,
)
from web3.providers.base import (
    BaseProvider,
)


def _get_block_by_something(method, params):
    block_identifier = params[0]
    if (
        block_identifier == 'latest' or
        block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000005'
    ):
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000005',
            'number': 5,
            'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000004',
            'transactions': [
                {'gasPrice': 70},
                {'gasPrice': 60},
                {'gasPrice': 60},
                {'gasPrice': 60},
                {'gasPrice': 15},
                {'gasPrice': 5},
                {'gasPrice': 50},
            ],
            'miner': '0xA',
            'timestamp': 100,
        }
    elif block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000004':
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000004',
            'number': 4,
            'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000003',
            'transactions': [
                {'gasPrice': 100},
                {'gasPrice': 80},
                {'gasPrice': 60},
            ],
            'miner': '0xB',
            'timestamp': 80,
        }
    elif block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000003':
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000003',
            'number': 3,
            'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000002',
            'transactions': [
                {'gasPrice': 100},
            ],
            'miner': '0xC',
            'timestamp': 60,
        }
    elif block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000002':
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000002',
            'number': 2,
            'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000001',
            'transactions': [
            ],
            'miner': '0xB',
            'timestamp': 40,
        }
    elif block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000001':
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000001',
            'number': 1,
            'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000000',
            'transactions': [
                {'gasPrice': 30},
                {'gasPrice': 35},
                {'gasPrice': 65},
            ],
            'miner': '0xA',
            'timestamp': 20,
        }
    elif (
        block_identifier == '0x0000000000000000000000000000000000000000000000000000000000000000' or
        block_identifier == 0
    ):
        return {
            'hash': '0x0000000000000000000000000000000000000000000000000000000000000000',
            'number': 0,
            'parentHash': None,
            'transactions': [
                {'gasPrice': 30},
                {'gasPrice': 50},
                {'gasPrice': 60},
                {'gasPrice': 30},
                {'gasPrice': 50},
                {'gasPrice': 60},
                {'gasPrice': 30},
                {'gasPrice': 50},
                {'gasPrice': 60},
                {'gasPrice': 30},
                {'gasPrice': 54},
                {'gasPrice': 10000000000000000000000},
            ],
            'miner': '0xA',
            'timestamp': 0,
        }
    else:
        assert False


@pytest.mark.parametrize(
    'strategy_params,expected',
    (
        # 120 second wait times
        (dict(max_wait_seconds=80, sample_size=5, probability=98), 70),
        (dict(max_wait_seconds=80, sample_size=5, probability=90), 25),
        (dict(max_wait_seconds=80, sample_size=5, probability=50), 11),
        # 60 second wait times
        (dict(max_wait_seconds=60, sample_size=5, probability=98), 92),
        (dict(max_wait_seconds=60, sample_size=5, probability=90), 49),
        (dict(max_wait_seconds=60, sample_size=5, probability=50), 11),
        # 40 second wait times
        (dict(max_wait_seconds=40, sample_size=5, probability=98), 100),
        (dict(max_wait_seconds=40, sample_size=5, probability=90), 81),
        (dict(max_wait_seconds=40, sample_size=5, probability=50), 11),
        # 20 second wait times
        (dict(max_wait_seconds=20, sample_size=5, probability=98), 100),
        (dict(max_wait_seconds=20, sample_size=5, probability=90), 100),
        (dict(max_wait_seconds=20, sample_size=5, probability=50), 36),
    ),
)
def test_time_based_gas_price_strategy(strategy_params, expected):
    fixture_middleware = construct_result_generator_middleware({
        'eth_getBlockByHash': _get_block_by_something,
        'eth_getBlockByNumber': _get_block_by_something,
    })

    w3 = Web3(
        providers=[BaseProvider()],
        middlewares=[fixture_middleware],
    )

    time_based_gas_price_strategy = construct_time_based_gas_price_strategy(
        **strategy_params,
    )
    w3.eth.setGasPriceStrategy(time_based_gas_price_strategy)
    actual = w3.eth.generateGasPrice()
    assert int(actual) == expected


@pytest.mark.parametrize(
    'strategy_params_zero,expected_exception_message',
    (
        # 120 second wait times, 0 sample_size
        (dict(max_wait_seconds=80, sample_size=0, probability=98),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=80, sample_size=0, probability=90),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=80, sample_size=0, probability=50),
            'Constrained sample size is 0'),
        # 60 second wait times, 0 sample_size
        (dict(max_wait_seconds=60, sample_size=0, probability=98),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=60, sample_size=0, probability=90),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=60, sample_size=0, probability=50),
            'Constrained sample size is 0'),
        # 40 second wait times, 0 sample_size
        (dict(max_wait_seconds=40, sample_size=0, probability=98),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=40, sample_size=0, probability=90),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=40, sample_size=0, probability=50),
            'Constrained sample size is 0'),
        # 20 second wait times, 0 sample_size
        (dict(max_wait_seconds=20, sample_size=0, probability=98),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=20, sample_size=0, probability=90),
            'Constrained sample size is 0'),
        (dict(max_wait_seconds=20, sample_size=0, probability=50),
            'Constrained sample size is 0'),
    ),
)
def test_time_based_gas_price_strategy_zero_sample(strategy_params_zero,
                                                   expected_exception_message):
    with pytest.raises(ValidationError) as excinfo:
        fixture_middleware = construct_result_generator_middleware({
            'eth_getBlockByHash': _get_block_by_something,
            'eth_getBlockByNumber': _get_block_by_something,
        })

        w3 = Web3(
            providers=[BaseProvider()],
            middlewares=[fixture_middleware],
        )
        time_based_gas_price_strategy_zero = construct_time_based_gas_price_strategy(
            **strategy_params_zero,
        )
        w3.eth.setGasPriceStrategy(time_based_gas_price_strategy_zero)
        w3.eth.generateGasPrice()
    assert str(excinfo.value) == expected_exception_message



def test_shh_asymmetric_key_pair(web3, skip_if_testrpc):
    skip_if_testrpc(web3)

    # Test generating key
    key_id = web3.shh.newKeyPair()
    assert web3.shh.hasKeyPair(key_id)
    assert len(web3.shh.getPublicKey(key_id)) == 132

    private_key = web3.shh.getPrivateKey(key_id)
    assert len(private_key) == 66
    assert web3.shh.deleteKeyPair(key_id)

    # Test adding a key
    assert not web3.shh.hasKeyPair(key_id)
    key_id = web3.shh.addPrivateKey(private_key)
    assert web3.shh.hasKeyPair(key_id)
    assert web3.shh.deleteKeyPair(key_id)


def test_shh_symmetric_key_pair(web3, skip_if_testrpc):
    skip_if_testrpc(web3)

    # Test generating key
    key_id = web3.shh.newSymKey()
    assert web3.shh.hasSymKey(key_id)

    key = web3.shh.getSymKey(key_id)
    assert len(key) == 66
    assert web3.shh.deleteSymKey(key_id)

    # Test adding a key
    assert not web3.shh.hasSymKey(key_id)
    key_id = web3.shh.addSymKey(key)
    assert web3.shh.hasSymKey(key_id)
    assert web3.shh.deleteSymKey(key_id)


def test_shh_symmetric_key_pair_from_password(web3, skip_if_testrpc):
    skip_if_testrpc(web3)

    key_id = web3.shh.generateSymKeyFromPassword('shh be quiet')

    assert web3.shh.hasSymKey(key_id)
    assert len(web3.shh.getSymKey(key_id)) == 66
    assert web3.shh.deleteSymKey(key_id)



from web3.contract import (
    parse_block_identifier_int,
)


#  This tests negative block number identifiers, which behave like python
#  list slices, with -1 being the latest block and -2 being the block before that.
#  This test is necessary because transaction calls allow negative block indexes, although
#  getBlock() does not allow negative block identifiers. Support for negative block identifier
#  will likely be removed in v5.
def test_parse_block_identifier_int(web3):
    last_num = web3.eth.getBlock('latest').number
    assert 0 == parse_block_identifier_int(web3, -1 - last_num)



# coding=utf-8

import pytest

from hexbytes import (
    HexBytes,
)

from web3 import Web3


@pytest.mark.parametrize(
    'message, digest',
    [
        ('cowmö', HexBytes('0x0f355f04c0a06eebac1d219b34c598f85a1169badee164be8a30345944885fe8')),
        ('', HexBytes('0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470')),
    ],
)
def test_sha3_text(message, digest):
    assert Web3.sha3(text=message) == digest


@pytest.mark.parametrize(
    'hexstr, digest',
    [
        (
            '0x636f776dc3b6',
            HexBytes('0x0f355f04c0a06eebac1d219b34c598f85a1169badee164be8a30345944885fe8')
        ),
        (
            '636f776dc3b6',
            HexBytes('0x0f355f04c0a06eebac1d219b34c598f85a1169badee164be8a30345944885fe8')
        ),
        (
            '0x',
            HexBytes('0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470')
        ),
        (
            '',
            HexBytes('0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470')
        ),
    ],
)
def test_sha3_hexstr(hexstr, digest):
    assert Web3.sha3(hexstr=hexstr) == digest


@pytest.mark.parametrize(
    'primitive, exception',
    [
        ('0x0', TypeError),
        ('', TypeError),
        (-1, ValueError),
    ],
)
def test_sha3_primitive_invalid(primitive, exception):
    with pytest.raises(exception):
        Web3.sha3(primitive)


@pytest.mark.parametrize(
    'primitive, digest',
    [
        (
            b'cowm\xc3\xb6',
            HexBytes('0x0f355f04c0a06eebac1d219b34c598f85a1169badee164be8a30345944885fe8')
        ),
        (
            b'',
            HexBytes('0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470')
        ),
    ],
)
def test_sha3_primitive(primitive, digest):
    assert Web3.sha3(primitive) == digest


@pytest.mark.parametrize(
    'kwargs',
    [
        {'text': ''},
        {'hexstr': '0x'},
        {'text': '', 'hexstr': '0x'},
    ],
)
def test_sha3_raise_if_primitive_and(kwargs):
    # must not set more than one input
    with pytest.raises(TypeError):
        Web3.sha3('', **kwargs)


def test_sha3_raise_if_hexstr_and_text():
    with pytest.raises(TypeError):
        Web3.sha3(hexstr='0x', text='')


def test_sha3_raise_if_no_args():
    with pytest.raises(TypeError):
        Web3.sha3()



import os
import pytest
import tempfile

from web3 import Web3

from .common import (
    GoEthereumEthModuleTest,
    GoEthereumNetModuleTest,
    GoEthereumPersonalModuleTest,
    GoEthereumTest,
    GoEthereumVersionModuleTest,
)
from .utils import (
    get_open_port,
    wait_for_socket,
)


@pytest.fixture(scope='module')
def geth_command_arguments(geth_binary, datadir, geth_ipc_path):
    geth_port = get_open_port()
    return (
        geth_binary,
        '--datadir', str(datadir),
        '--ipcpath', geth_ipc_path,
        '--nodiscover',
        '--fakepow',
        '--port', geth_port,
    )


@pytest.fixture(scope='module')
def geth_ipc_path(datadir):
    geth_ipc_dir_path = tempfile.mkdtemp()
    _geth_ipc_path = os.path.join(geth_ipc_dir_path, 'geth.ipc')
    yield _geth_ipc_path

    if os.path.exists(_geth_ipc_path):
        os.remove(_geth_ipc_path)


@pytest.fixture(scope="module")
def web3(geth_process, geth_ipc_path):
    wait_for_socket(geth_ipc_path)
    _web3 = Web3(Web3.IPCProvider(geth_ipc_path))
    return _web3


class TestGoEthereumTest(GoEthereumTest):
    pass


class TestGoEthereumEthModuleTest(GoEthereumEthModuleTest):
    pass


class TestGoEthereumVersionModuleTest(GoEthereumVersionModuleTest):
    pass


class TestGoEthereumNetModuleTest(GoEthereumNetModuleTest):
    pass


class TestGoEthereumPersonalModuleTest(GoEthereumPersonalModuleTest):
    pass



# coding=utf-8

import pytest

from web3 import Web3


@pytest.mark.parametrize(
    'val, expected',
    (
        (b'\x01', b'\x01'),
        (b'\xff', b'\xff'),
        (b'\x00', b'\x00'),
        (0x1, b'\x01'),
        (0x0001, b'\x01'),
        (0xFF, b'\xff'),
        (0, b'\x00'),
        (256, b'\x01\x00'),
        (True, b'\x01'),
        (False, b'\x00'),
    ),
)
def test_to_bytes_primitive(val, expected):
    assert Web3.toBytes(val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('0x', b''),
        ('0x0', b'\x00'),
        ('0x1', b'\x01'),
        ('0', b'\x00'),
        ('1', b'\x01'),
        ('0xFF', b'\xff'),
        ('0x100', b'\x01\x00'),
        ('0x0000', b'\x00\x00'),
        ('0000', b'\x00\x00'),
    ),
)
def test_to_bytes_hexstr(val, expected):
    assert Web3.toBytes(hexstr=val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('cowmö', b'cowm\xc3\xb6'),
        ('', b''),
    ),
)
def test_to_bytes_text(val, expected):
    assert Web3.toBytes(text=val) == expected


def test_to_text_identity():
    assert Web3.toText(text='pass-through') == 'pass-through'


@pytest.mark.parametrize(
    'val, expected',
    (
        (b'', ''),
        ('0x', ''),
        (b'cowm\xc3\xb6', 'cowmö'),
        ('0x636f776dc3b6', 'cowmö'),
        (0x636f776dc3b6, 'cowmö'),
        ('0xa', '\n'),
    ),
)
def test_to_text(val, expected):
    assert Web3.toText(val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('0x', ''),
        ('0xa', '\n'),
        ('0x636f776dc3b6', 'cowmö'),
        ('636f776dc3b6', 'cowmö'),
    ),
)
def test_to_text_hexstr(val, expected):
    assert Web3.toText(hexstr=val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        (b'\x00', 0),
        (b'\x01', 1),
        (b'\x00\x01', 1),
        (b'\x01\x00', 256),
        (True, 1),
        (False, 0),
        ('255', TypeError),
        ('-1', TypeError),
        ('0x0', TypeError),
        ('0x1', TypeError),
    ),
)
def test_to_int(val, expected):
    if isinstance(expected, type):
        with pytest.raises(expected):
            Web3.toInt(val)
    else:
        assert Web3.toInt(val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('0', 0),
        ('-1', -1),
        ('255', 255),
        ('0x0', ValueError),
        ('0x1', ValueError),
        ('1.1', ValueError),
        ('a', ValueError),
    ),
)
def test_to_int_text(val, expected):
    if isinstance(expected, type):
        with pytest.raises(expected):
            Web3.toInt(text=val)
    else:
        assert Web3.toInt(text=val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('0x0', 0),
        ('0x1', 1),
        ('0x01', 1),
        ('0x10', 16),
        ('0', 0),
        ('1', 1),
        ('01', 1),
        ('10', 16),
    ),
)
def test_to_int_hexstr(val, expected):
    assert Web3.toInt(hexstr=val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        (b'\x00', '0x00'),
        (b'\x01', '0x01'),
        (b'\x10', '0x10'),
        (b'\x01\x00', '0x0100'),
        (b'\x00\x0F', '0x000f'),
        (b'', '0x'),
        (0, '0x0'),
        (1, '0x1'),
        (16, '0x10'),
        (256, '0x100'),
        (0x0, '0x0'),
        (0x0F, '0xf'),
        (False, '0x0'),
        (True, '0x1'),
    ),
)
def test_to_hex(val, expected):
    assert Web3.toHex(val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('', '0x'),
        ('cowmö', '0x636f776dc3b6'),
    ),
)
def test_to_hex_text(val, expected):
    assert Web3.toHex(text=val) == expected


@pytest.mark.parametrize(
    'val, expected',
    (
        ('0x0', '0x0'),
        ('0x1', '0x1'),
        ('0x0001', '0x0001'),
        ('0x10', '0x10'),
        ('0xF', '0xf'),
        ('F', '0xf'),
    ),
)
def test_to_hex_cleanup_only(val, expected):
    assert Web3.toHex(hexstr=val) == expected




import json
import pytest

from eth_tester import (
    EthereumTester,
)

from ens import ENS
from ens.contract_data import (
    registrar_abi,
    registrar_bytecode,
    registrar_bytecode_runtime,
    resolver_abi,
    resolver_bytecode,
    resolver_bytecode_runtime,
    reverse_registrar_abi,
    reverse_registrar_bytecode,
    reverse_registrar_bytecode_runtime,
    reverse_resolver_abi,
    reverse_resolver_bytecode,
    reverse_resolver_bytecode_runtime,
)
from web3 import Web3
from web3.contract import (
    Contract,
)
from web3.providers.eth_tester import (
    EthereumTesterProvider,
)


def bytes32(val):
    if isinstance(val, int):
        result = Web3.toBytes(val)
    else:
        raise TypeError('val %r could not be converted to bytes')
    if len(result) < 32:
        return result.rjust(32, b'\0')
    else:
        return result


def deploy(w3, Factory, from_address, args=None):
    args = args or []
    factory = Factory(w3)
    deploy_txn = factory.constructor(*args).transact({'from': from_address})
    deploy_receipt = w3.eth.waitForTransactionReceipt(deploy_txn)
    assert deploy_receipt is not None
    return factory(address=deploy_receipt['contractAddress'])


def DefaultReverseResolver(w3):
    return w3.eth.contract(
        bytecode=reverse_resolver_bytecode,
        bytecode_runtime=reverse_resolver_bytecode_runtime,
        abi=reverse_resolver_abi,
        ContractFactoryClass=Contract,
    )


def ReverseRegistrar(w3):
    return w3.eth.contract(
        bytecode=reverse_registrar_bytecode,
        bytecode_runtime=reverse_registrar_bytecode_runtime,
        abi=reverse_registrar_abi,
        ContractFactoryClass=Contract,
    )


def PublicResolverFactory(w3):
    return w3.eth.contract(
        bytecode=resolver_bytecode,
        bytecode_runtime=resolver_bytecode_runtime,
        abi=resolver_abi,
        ContractFactoryClass=Contract,
    )


def ENSFactory(w3):
    return w3.eth.contract(
        bytecode="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",  # noqa: E501
        bytecode_runtime="606060405236156100805763ffffffff7c01000000000000000000000000000000000000000000000000000000006000350416630178b8bf811461008557806302571be3146100b757806306ab5923146100cd57806314ab9038146100f457806316a25cbd146101175780631896f70a1461014a5780635b0fc9c31461016c575b600080fd5b341561009057600080fd5b61009b60043561018e565b604051600160a060020a03909116815260200160405180910390f35b34156100c257600080fd5b61009b6004356101ac565b34156100d857600080fd5b6100f2600435602435600160a060020a03604435166101c7565b005b34156100ff57600080fd5b6100f260043567ffffffffffffffff60243516610289565b341561012257600080fd5b61012d600435610355565b60405167ffffffffffffffff909116815260200160405180910390f35b341561015557600080fd5b6100f2600435600160a060020a036024351661038c565b341561017757600080fd5b6100f2600435600160a060020a0360243516610432565b600090815260208190526040902060010154600160a060020a031690565b600090815260208190526040902054600160a060020a031690565b600083815260208190526040812054849033600160a060020a039081169116146101f057600080fd5b8484604051918252602082015260409081019051908190039020915083857fce0457fe73731f824cc272376169235128c118b49d344817417c6d108d155e8285604051600160a060020a03909116815260200160405180910390a3506000908152602081905260409020805473ffffffffffffffffffffffffffffffffffffffff1916600160a060020a03929092169190911790555050565b600082815260208190526040902054829033600160a060020a039081169116146102b257600080fd5b827f1d4f9bbfc9cab89d66e1a1562f2233ccbf1308cb4f63de2ead5787adddb8fa688360405167ffffffffffffffff909116815260200160405180910390a250600091825260208290526040909120600101805467ffffffffffffffff90921674010000000000000000000000000000000000000000027fffffffff0000000000000000ffffffffffffffffffffffffffffffffffffffff909216919091179055565b60009081526020819052604090206001015474010000000000000000000000000000000000000000900467ffffffffffffffff1690565b600082815260208190526040902054829033600160a060020a039081169116146103b557600080fd5b827f335721b01866dc23fbee8b6b2c7b1e14d6f05c28cd35a2c934239f94095602a083604051600160a060020a03909116815260200160405180910390a250600091825260208290526040909120600101805473ffffffffffffffffffffffffffffffffffffffff1916600160a060020a03909216919091179055565b600082815260208190526040902054829033600160a060020a0390811691161461045b57600080fd5b827fd4735d920b0f87494915f556dd9b54c8f309026070caea5c737245152564d26683604051600160a060020a03909116815260200160405180910390a250600091825260208290526040909120805473ffffffffffffffffffffffffffffffffffffffff1916600160a060020a039092169190911790555600a165627a7a7230582050975b6c54a16d216b563f4c4960d6ebc5881eb1ec73c2ef1f87920a251159530029",  # noqa: E501
        abi=json.loads('[{"constant":true,"inputs":[{"name":"node","type":"bytes32"}],"name":"resolver","outputs":[{"name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"node","type":"bytes32"}],"name":"owner","outputs":[{"name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"node","type":"bytes32"},{"name":"label","type":"bytes32"},{"name":"owner","type":"address"}],"name":"setSubnodeOwner","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"node","type":"bytes32"},{"name":"ttl","type":"uint64"}],"name":"setTTL","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"name":"node","type":"bytes32"}],"name":"ttl","outputs":[{"name":"","type":"uint64"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"node","type":"bytes32"},{"name":"resolver","type":"address"}],"name":"setResolver","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"node","type":"bytes32"},{"name":"owner","type":"address"}],"name":"setOwner","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"inputs":[],"payable":false,"stateMutability":"nonpayable","type":"constructor"},{"anonymous":false,"inputs":[{"indexed":true,"name":"node","type":"bytes32"},{"indexed":true,"name":"label","type":"bytes32"},{"indexed":false,"name":"owner","type":"address"}],"name":"NewOwner","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"node","type":"bytes32"},{"indexed":false,"name":"owner","type":"address"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"node","type":"bytes32"},{"indexed":false,"name":"resolver","type":"address"}],"name":"NewResolver","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"node","type":"bytes32"},{"indexed":false,"name":"ttl","type":"uint64"}],"name":"NewTTL","type":"event"}]'),  # noqa: E501
        ContractFactoryClass=Contract,
    )


def ETHRegistrarFactory(w3):
    return w3.eth.contract(
        bytecode=registrar_bytecode,
        bytecode_runtime=registrar_bytecode_runtime,
        abi=registrar_abi,
        ContractFactoryClass=Contract,
    )


# session scope for performance
@pytest.fixture(scope="session")
def ens_setup():
    w3 = Web3(EthereumTesterProvider(EthereumTester()))

    # ** Set up ENS contracts **

    # remove account that creates ENS, so test transactions don't have write access
    accounts = w3.eth.accounts
    ens_key = accounts.pop()

    # create ENS contract
    eth_labelhash = w3.sha3(text='eth')
    eth_namehash = bytes32(0x93cdeb708b7545dc668eb9280176169d1c33cfd8ed6f04690a0bcc88a93fc4ae)
    resolver_namehash = bytes32(0xfdd5d5de6dd63db72bbc2d487944ba13bf775b50a80805fe6fcaba9b0fba88f5)
    reverse_tld_namehash = bytes32(0xa097f6721ce401e757d1223a763fef49b8b5f90bb18567ddb86fd205dff71d34)  # noqa: E501
    reverser_namehash = bytes32(0x91d1777781884d03a6757a803996e38de2a42967fb37eeaca72729271025a9e2)
    ens_contract = deploy(w3, ENSFactory, ens_key)

    # create public resolver
    public_resolver = deploy(w3, PublicResolverFactory, ens_key, args=[ens_contract.address])

    # set 'resolver.eth' to resolve to public resolver
    ens_contract.functions.setSubnodeOwner(
        b'\0' * 32,
        eth_labelhash,
        ens_key
    ).transact({'from': ens_key})

    ens_contract.functions.setSubnodeOwner(
        eth_namehash,
        w3.sha3(text='resolver'),
        ens_key
    ).transact({'from': ens_key})

    ens_contract.functions.setResolver(
        resolver_namehash,
        public_resolver.address
    ).transact({'from': ens_key})

    public_resolver.functions.setAddr(
        resolver_namehash,
        public_resolver.address
    ).transact({'from': ens_key})

    # create .eth auction registrar
    eth_registrar = deploy(
        w3,
        ETHRegistrarFactory,
        ens_key,
        args=[ens_contract.address, eth_namehash, 1],
    )

    # set '.eth' to resolve to the registrar
    ens_contract.functions.setResolver(
        eth_namehash,
        public_resolver.address
    ).transact({'from': ens_key})

    public_resolver.functions.setAddr(
        eth_namehash,
        eth_registrar.address
    ).transact({'from': ens_key})

    # create reverse resolver
    reverse_resolver = deploy(w3, DefaultReverseResolver, ens_key, args=[ens_contract.address])

    # create reverse registrar
    reverse_registrar = deploy(
        w3,
        ReverseRegistrar,
        ens_key,
        args=[ens_contract.address, reverse_resolver.address]
    )

    # set 'addr.reverse' to resolve to reverse registrar
    ens_contract.functions.setSubnodeOwner(
        b'\0' * 32,
        w3.sha3(text='reverse'),
        ens_key
    ).transact({'from': ens_key})

    ens_contract.functions.setSubnodeOwner(
        reverse_tld_namehash,
        w3.sha3(text='addr'),
        ens_key
    ).transact({'from': ens_key})

    ens_contract.functions.setResolver(
        reverser_namehash,
        public_resolver.address
    ).transact({'from': ens_key})

    public_resolver.functions.setAddr(
        reverser_namehash,
        reverse_registrar.address
    ).transact({'from': ens_key})

    # set owner of tester.eth to an account controlled by tests
    ens_contract.functions.setSubnodeOwner(
        eth_namehash,
        w3.sha3(text='tester'),
        w3.eth.accounts[2]  # note that this does not have to be the default, only in the list
    ).transact({'from': ens_key})

    # make the registrar the owner of the 'eth' name
    ens_contract.functions.setSubnodeOwner(
        b'\0' * 32,
        eth_labelhash,
        eth_registrar.address
    ).transact({'from': ens_key})

    # make the reverse registrar the owner of the 'addr.reverse' name
    ens_contract.functions.setSubnodeOwner(
        reverse_tld_namehash,
        w3.sha3(text='addr'),
        reverse_registrar.address
    ).transact({'from': ens_key})

    return ENS.fromWeb3(w3, ens_contract.address)


@pytest.fixture
def ens(ens_setup, mocker):
    mocker.patch('web3.middleware.stalecheck._isfresh', return_value=True)
    ens_setup.web3.eth.defaultAccount = ens_setup.web3.eth.coinbase
    return ens_setup


@pytest.fixture()
def TEST_ADDRESS(address_conversion_func):
    return address_conversion_func("0x000000000000000000000000000000000000dEaD")
